"""
The *glyphNameComboBox* submodule
-----------------------------

The *glyphNameComboBox* submodule provides a QComboBox_ widget that performs
autocompletion from a Font_’s glyph names.

.. _Font: http://ts-defcon.readthedocs.org/en/ufo3/objects/font.html
.. _QComboBox: http://doc.qt.io/qt-5/qcombobox.html
"""

from PyQt5.QtCore import QStringListModel
from PyQt5.QtWidgets import QComboBox, QCompleter

from defconQt.tools.textSplitter import splitText

__all__ = ["GlyphNameComboBox"]


class GlyphNameCompleter(QCompleter):
    def __init__(self, font, parent=None):
        super().__init__(parent)
        self._font = font
        self.setCompletionMode(QCompleter.InlineCompletion)

    def splitPath(self, path):
        # hack around the splitPath() function to feed custom results to the
        # QCompleter, see: http://stackoverflow.com/a/28286322/2037879
        match, replace = _search(path, self._font)
        if replace:
            comboBox = self.widget()
            comboBox.setCurrentText(match)
            comboBox.lineEdit().selectAll()
            match = None
        if match is None:
            model = QStringListModel()
        else:
            model = QStringListModel([match])
        self.setModel(model)
        return [path]


class GlyphNameComboBox(QComboBox):
    """
    # TODO: consider popup completer or adding font glyph names?
    """

    splitTextFunction = splitText

    def __init__(self, font, parent=None):
        super().__init__(parent)
        self.setEditable(True)
        completer = GlyphNameCompleter(font)
        self.setCompleter(completer)


def _search(text, font):
    # no text
    if not text:
        return None, False
    glyphNames = font.keys()
    match = None
    # direct match
    if text in glyphNames:
        match = text
    # character entry
    elif len(text) == 1:
        uniValue = ord(text)
        gName = font.unicodeData.glyphNameForUnicode(uniValue)
        if gName is not None:
            return gName, True
    # fallback. find closest match
    if not match:
        for glyphName in sorted(glyphNames):
            if glyphName.startswith(text):
                match = glyphName
                break
    return match, False



import os
import re

from PyQt5.QtCore import QEvent, QSize, QSizeF, QStandardPaths, Qt, pyqtSignal
from PyQt5.QtGui import (
    QBrush,
    QColor,
    QCursor,
    QIcon,
    QIntValidator,
    QPainter,
    QPalette,
)
from PyQt5.QtPrintSupport import QPrinter
from PyQt5.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QComboBox,
    QMenu,
    QPushButton,
    QSizePolicy,
    QSlider,
    QStyledItemDelegate,
    QTableWidget,
    QTableWidgetItem,
    QToolBar,
    QToolTip,
    QVBoxLayout,
    QWidgetAction,
)

from defconQt.controls.glyphLineView import GlyphLineView, GlyphLineWidget
from defconQt.controls.glyphSequenceEdit import GlyphSequenceComboBox, GlyphSequenceEdit
from defconQt.windows.baseWindows import BaseWindow
from trufont.objects import settings
from trufont.objects.defcon import TGlyph
from trufont.resources import icons_db  # noqa

pointSizes = [50, 75, 100, 125, 150, 200, 250, 300, 350, 400, 450, 500]


class MetricsWindow(BaseWindow):
    def __init__(self, font, string=None, parent=None):
        super().__init__(parent, Qt.Window)
        self.setAttribute(Qt.WA_DeleteOnClose, False)

        if string is None:
            try:
                import getpass

                string = getpass.getuser()
            except Exception:
                string = self.tr("World")
            string = self.tr("Hello {0}").format(string)

        self.toolbar = MetricsToolBar(font, self)
        self.lineView = MetricsLineView(self)
        self.table = MetricsTable(self)

        self.toolbar.glyphsChanged.connect(self.lineView.setGlyphRecords)
        self.toolbar.glyphsChanged.connect(self.table.setGlyphs)
        self.toolbar.pointSizeChanged.connect(self.lineView.setPointSize)
        self.toolbar.settingsChanged.connect(self.lineView.setSettings)
        self.toolbar.settingsChanged.connect(
            lambda s: self.table.setKerningEnabled(s["showKerning"])
        )
        self.lineView.glyphActivated.connect(self._glyphActivated)
        self.lineView.pointSizeModified.connect(self.toolbar.setPointSize)
        self.lineView.selectionModified.connect(self.table.setCurrentGlyph)
        self.table.selectedIndexChanged.connect(self.lineView.setSelected)

        self.toolbar.setPointSize(self.lineView.pointSize())
        self.toolbar.setWrapLines(True)
        self.toolbar.setText(string)

        layout = QVBoxLayout(self)
        layout.addWidget(self.toolbar)
        layout.addWidget(self.lineView)
        layout.addWidget(self.table)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        self.setLayout(layout)

        self.font = font
        self.font.info.addObserver(self, "_fontInfoChanged", "Info.Changed")

        self.updateWindowTitle(font=font)

        self.readSettings()

    def readSettings(self):
        geometry = settings.metricsWindowGeometry()
        if geometry:
            self.restoreGeometry(geometry)

    def writeSettings(self):
        settings.setMetricsWindowGeometry(self.saveGeometry())

    def updateWindowTitle(self, title=None, font=None):
        if title is None:
            title = self.tr("Metrics")
        if font is not None:
            title = "{} – {} {}".format(
                title, font.info.familyName, font.info.styleName
            )
        self.setWindowTitle(title)

    def setGlyphs(self, glyphs):
        glyphNames = []
        for glyph in glyphs:
            if glyph.unicode is not None:
                glyphNames.append(chr(glyph.unicode))
            else:
                glyphNames.append("/%s " % glyph.name)
        self.setText("".join(glyphNames))

    def setText(self, text):
        self.toolbar.setText(text)

    # -------------
    # Notifications
    # -------------

    # font

    def _fontInfoChanged(self, notification):
        self.updateWindowTitle(font=self.font)

    # widget

    def _glyphActivated(self, glyph):
        fontWindow = self.parent() or self._fontWindow
        if fontWindow is not None:
            fontWindow.openGlyphTab(glyph)
            fontWindow.raise_()

    # ----------
    # Qt methods
    # ----------

    def sizeHint(self):
        return QSize(1150, 630)

    def moveEvent(self, event):
        self.writeSettings()

    resizeEvent = moveEvent

    def showEvent(self, event):
        app = QApplication.instance()
        data = dict(window=self)
        app.postNotification("metricsWindowWillOpen", data)
        super().showEvent(event)
        app.postNotification("metricsWindowOpened", data)

    def closeEvent(self, event):
        super().closeEvent(event)
        if event.isAccepted():
            app = QApplication.instance()
            data = dict(window=self)
            app.postNotification("metricsWindowWillClose", data)
            self.font.info.removeObserver(self, "Info.Changed")
            self.toolbar.closeEvent(event)
            self.lineView.closeEvent(event)


# -------------------
# Text to glyph names
# -------------------

escapeRep = {"//": "/slash ", "\\n": "\u2029"}
escapeRep = {re.escape(k): v for k, v in escapeRep.items()}
escapeRe = re.compile("|".join(escapeRep.keys()))


def _cmapFunc(c, unicodeData):
    uni = ord(c)
    if uni == 0x2029:
        glyphName = c
    else:
        glyphName = unicodeData.glyphNameForUnicode(uni)
    return glyphName


def _stackCompileFunc(glyphNames, stack):
    if stack == ["?"]:
        app = QApplication.instance()
        glyph = app.currentGlyph()
        if glyph is not None:
            glyphNames.append(glyph.name)
    elif stack:
        glyphNames.append("".join(stack))


def _textEscapeFunc(text):
    return escapeRe.sub(lambda m: escapeRep[re.escape(m.group(0))], text)


def _glyphs(self):
    text = self.text()
    glyphNames = self.splitTextFunction(
        text,
        self._font.unicodeData,
        cmapFunc=_cmapFunc,
        compileFunc=_stackCompileFunc,
        escapeFunc=_textEscapeFunc,
    )
    glyphs = []
    for glyphName in glyphNames:
        if glyphName == "\u2029":
            glyph = TGlyph()
            glyph.unicode = 2029
            glyphs.append(glyph)
        elif glyphName in self._font:
            glyphs.append(self._font[glyphName])
    return glyphs


class MetricsSequenceComboBox(GlyphSequenceComboBox):
    glyphs = _glyphs


class MetricsSequenceEdit(GlyphSequenceEdit):
    glyphs = _glyphs


class MetricsToolBar(QToolBar):
    """
    Emits *pointSizeChanged*.
    """

    glyphsChanged = pyqtSignal(list)
    settingsChanged = pyqtSignal(dict)

    def __init__(self, font, parent=None):
        super().__init__(parent)
        auxiliaryWidth = self.fontMetrics().width("0") * 8
        self.leftTextField = MetricsSequenceEdit(font, self)
        self.leftTextField.setMaximumWidth(auxiliaryWidth)
        self.textField = MetricsSequenceComboBox(font, self)
        # XXX: had to use Maximum because Preferred did extend the widget(?)
        self.textField.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum)
        self.rightTextField = MetricsSequenceEdit(font, self)
        self.rightTextField.setMaximumWidth(auxiliaryWidth)
        self.leftTextField.textEdited.connect(self.textField.editTextChanged)
        self.rightTextField.textEdited.connect(self.textField.editTextChanged)
        self.textField.editTextChanged.connect(self._textChanged)

        self.comboBox = QComboBox(self)
        self.comboBox.setEditable(True)
        self.comboBox.setCompleter(None)
        self.comboBox.setValidator(QIntValidator(self))
        for p in pointSizes:
            self.comboBox.addItem(str(p))
        self.pointSizeChanged = self.comboBox.currentIndexChanged[str]

        self.configBar = QPushButton(self)
        self.configBar.setFlat(True)
        self.configBar.setIcon(QIcon(":settings.svg"))
        self.configBar.setStyleSheet("padding: 2px 0px; padding-right: 10px")
        self.toolsMenu = QMenu(self)
        self._showKerning = self.toolsMenu.addAction(
            self.tr("Show Kerning"), self._kerningVisibilityChanged
        )
        self._showKerning.setCheckable(True)
        self._showMetrics = self.toolsMenu.addAction(
            self.tr("Show Metrics"), self._controlsTriggered
        )
        self._showMetrics.setCheckable(True)
        self.toolsMenu.addSeparator()
        self._verticalFlip = self.toolsMenu.addAction(
            self.tr("Vertical Flip"), self._controlsTriggered
        )
        self._verticalFlip.setCheckable(True)
        self._wrapLines = self.toolsMenu.addAction(
            self.tr("Wrap Lines"), self._controlsTriggered
        )
        self._wrapLines.setCheckable(True)
        self.toolsMenu.addSeparator()
        action = self.toolsMenu.addAction(self.tr("Line Height:"))
        action.setEnabled(False)
        lineHeight = QWidgetAction(self.toolsMenu)
        self._lineHeightSlider = slider = QSlider(Qt.Horizontal, self)
        # QSlider works with integers so we'll just divide what comes out of it
        # by 100
        slider.setMinimum(80)
        slider.setMaximum(160)
        slider.setValue(110)
        slider.valueChanged.connect(self._controlsTriggered)
        slider.valueChanged.connect(self._sliderLineHeightChanged)
        lineHeight.setDefaultWidget(slider)
        self.toolsMenu.addAction(lineHeight)
        self.configBar.setMenu(self.toolsMenu)

        self.addWidget(self.leftTextField)
        self.addWidget(self.textField)
        self.addWidget(self.rightTextField)
        self.addWidget(self.comboBox)
        self.addWidget(self.configBar)

        app = QApplication.instance()
        app.dispatcher.addObserver(self, "_currentGlyphChanged", "currentGlyphChanged")

        self.readSettings()

    def readSettings(self):
        items = settings.metricsWindowComboBoxItems()
        self.textField.clear()
        self.textField.addItems(items)

    # -------------
    # Notifications
    # -------------

    # app

    def _currentGlyphChanged(self, notification):
        self._textChanged()

    # widget

    def _controlsTriggered(self):
        params = dict(
            lineHeight=self._lineHeightSlider.value() / 100,
            showKerning=self._showKerning.isChecked(),
            showMetrics=self._showMetrics.isChecked(),
            verticalFlip=self._verticalFlip.isChecked(),
            wrapLines=self._wrapLines.isChecked(),
        )
        self.settingsChanged.emit(params)

    def _kerningVisibilityChanged(self):
        self._controlsTriggered()
        # if showKerning was triggered, it won't apply until we pipe the glyphs
        # again. do so
        self._textChanged()

    def _sliderLineHeightChanged(self, value):
        QToolTip.showText(QCursor.pos(), str(value / 100), self)

    def _textChanged(self):
        leftGlyphs = self.leftTextField.glyphs()
        rightGlyphs = self.rightTextField.glyphs()
        glyphs = self.textField.glyphs()
        ret = []
        for glyph in glyphs:
            ret.extend(leftGlyphs + [glyph] + rightGlyphs)
        self.glyphsChanged.emit(ret)

    # --------------
    # Public methods
    # --------------

    def setPointSize(self, pointSize):
        self.comboBox.blockSignals(True)
        self.comboBox.setEditText(str(pointSize))
        if pointSize in pointSizes:
            self.comboBox.setCurrentIndex(pointSizes.index(pointSize))
        self.comboBox.blockSignals(False)

    def setText(self, text, left=None, right=None):
        self.leftTextField.setText(left)
        self.rightTextField.setText(right)
        self.textField.setEditText(text)

    def setWrapLines(self, value):
        self._wrapLines.setChecked(value)
        self._controlsTriggered()

    # TODO: more methods

    # ----------
    # Qt methods
    # ----------

    def closeEvent(self, event):
        super().closeEvent(event)
        if event.isAccepted():
            app = QApplication.instance()
            app.dispatcher.removeObserver(self, "currentGlyphChanged")

    def showEvent(self, event):
        super().showEvent(event)
        self.textField.setFocus(Qt.OtherFocusReason)


class MetricsLineWidget(GlyphLineWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAttribute(Qt.WA_KeyCompression)

        # inbound notification
        app = QApplication.instance()
        app.dispatcher.addObserver(self, "_needsUpdate", "metricsViewUpdate")

    def drawGlyphForeground(self, painter, glyph, rect, selected=False):
        app = QApplication.instance()
        # TODO: no scale param with self._inverseScale? or getter function as
        # in glyphView?
        data = dict(widget=self, painter=painter, selected=selected)
        app.postNotification("metricsViewDraw", data)

    # -------------
    # Notifications
    # -------------

    def _needsUpdate(self, notification):
        self.update()

    # --------------
    # Public methods
    # --------------

    def exportToPDF(self, path=None):
        if path is None:
            desktop = QStandardPaths.standardLocations(QStandardPaths.DesktopLocation)
            path = os.path.join(desktop[0], "metricsWindow.pdf")

        printer = QPrinter(QPrinter.ScreenResolution)
        printer.setOutputFileName(path)
        printer.setOutputFormat(QPrinter.PdfFormat)
        printer.setFullPage(True)
        printer.setPaperSize(QSizeF(self.size()), QPrinter.DevicePixel)

        painter = QPainter()
        painter.begin(printer)
        painter.setRenderHint(QPainter.Antialiasing)
        if self._rightToLeft:
            self.paintRightToLeft(painter, self.geometry())
        else:
            self.paintLeftToRight(painter, self.geometry())
        painter.end()

    # ----------
    # Qt methods
    # ----------

    def closeEvent(self, event):
        super().closeEvent(event)
        if event.isAccepted():
            app = QApplication.instance()
            app.dispatcher.removeObserver(self, "metricsViewUpdate")

    def _arrowKeyPressEvent(self, event):
        key = event.key()
        modifiers = event.modifiers()
        if self._selected is not None:
            glyph = self._glyphRecords[self._selected].glyph
            # TODO: not really DRY w other widgets
            delta = event.count()
            if modifiers & Qt.ShiftModifier:
                delta *= 10
                if modifiers & Qt.ControlModifier:
                    delta *= 10
            if key == Qt.Key_Left:
                delta = -delta
            if modifiers & Qt.AltModifier:
                if glyph.leftMargin is not None:
                    glyph.leftMargin += delta
            else:
                glyph.width += delta

    def _navKeyPressEvent(self, event):
        key = event.key()
        if self._selected is not None:
            glyphCount = len(self._glyphRecords)
            delta = event.count()
            if key == Qt.Key_Home:
                delta = -delta
            newSelected = self._selected + delta
            if newSelected < 0 or newSelected >= glyphCount:
                return
            self.setSelected(newSelected)
            self.selectionModified.emit(self._selected)

    def keyPressEvent(self, event):
        key = event.key()
        if key in (Qt.Key_Home, Qt.Key_End):
            self._navKeyPressEvent(event)
        elif key in (Qt.Key_Left, Qt.Key_Right):
            self._arrowKeyPressEvent(event)
        else:
            super().keyPressEvent(event)
        app = QApplication.instance()
        data = dict(event=event, widget=self)
        app.postNotification("metricsViewKeyPress", data)

    def keyReleaseEvent(self, event):
        super().keyReleaseEvent(event)
        app = QApplication.instance()
        data = dict(event=event, widget=self)
        app.postNotification("metricsViewKeyRelease", data)


class MetricsLineView(GlyphLineView):
    glyphLineWidgetClass = MetricsLineWidget

    def exportToPDF(self, path=None):
        self._glyphLineWidget.exportToPDF(path)

    def setSettings(self, settings):
        self.setApplyKerning(settings["showKerning"])
        self.setLineHeight(settings["lineHeight"])
        self.setDrawMetrics(settings["showMetrics"])
        self.setVerticalFlip(settings["verticalFlip"])
        self.setWrapLines(settings["wrapLines"])

    def closeEvent(self, event):
        super().closeEvent(event)
        self._glyphLineWidget.closeEvent(event)


class MetricsTableItem(QTableWidgetItem):
    def setData(self, role, value):
        if role & Qt.EditRole:
            # don't set empty data
            # TODO: maybe fetch the value from cell back to the editor
            if value == "":
                return
        super().setData(role, value)


class MetricsTableItemDelegate(QStyledItemDelegate):
    def createEditor(self, parent, option, index):
        editor = super().createEditor(parent, option, index)
        # editor.setAlignment(Qt.AlignCenter)
        editor.setValidator(QIntValidator(self))
        return editor

    # TODO: implement =... lexer
    # TODO: Alt+left or Alt+right don't SelectAll of the new cell
    # cell by default. Implement this.
    # TODO: cycle b/w editable cell area
    def eventFilter(self, editor, event):
        if event.type() == QEvent.KeyPress:
            chg = None
            count = event.count()
            key = event.key()
            if key == Qt.Key_Up:
                chg = count
            elif key == Qt.Key_Down:
                chg = -count
            elif not key == Qt.Key_Return:
                return False
            if chg is not None:
                modifiers = event.modifiers()
                if modifiers & Qt.AltModifier:
                    return False
                elif modifiers & Qt.ShiftModifier:
                    chg *= 10
                    if modifiers & Qt.ControlModifier:
                        chg *= 10
                cur = int(editor.text())
                editor.setText(str(cur + chg))
            self.commitData.emit(editor)
            editor.selectAll()
            return True
        return False


class MetricsTable(QTableWidget):
    selectedIndexChanged = pyqtSignal(object)

    def __init__(self, parent=None):
        super().__init__(5, 1, parent)
        self.setAttribute(Qt.WA_KeyCompression)
        self.setItemDelegate(MetricsTableItemDelegate(self))
        data = [
            None,
            self.tr("Width"),
            self.tr("Left"),
            self.tr("Right"),
            self.tr("Kerning"),
        ]
        # Don't grey-out disabled cells
        palette = self.palette()
        fgColor = palette.color(QPalette.Text)
        palette.setColor(QPalette.Disabled, QPalette.Text, fgColor)
        self.setPalette(palette)
        for index, title in enumerate(data):
            item = MetricsTableItem(title)
            item.setFlags(Qt.NoItemFlags)
            item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
            self.setItem(index, 0, item)
        # let's use this one column to compute the width of others
        columnWidth = self.columnWidth(0)
        self._cellWidth = 0.5 * columnWidth
        self.setColumnWidth(0, 0.55 * columnWidth)
        self.horizontalHeader().hide()
        self.verticalHeader().hide()
        self._coloredColumn = None
        self._kerningEnabled = False

        # always show a scrollbar to fix layout
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        self._glyphs = []
        self.fillGlyphs()
        self.resizeRowsToContents()
        self.currentItemChanged.connect(self._itemChanged)
        self.cellChanged.connect(self._cellEdited)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        # edit cell on single click, not double
        self.setEditTriggers(QAbstractItemView.CurrentChanged)

    # -------------
    # Notifications
    # -------------

    # font

    def _glyphChanged(self, notification):
        self.updateCells()

    def _subscribeToGlyphs(self, glyphs):
        handledGlyphs = set()
        for glyph in glyphs:
            if glyph in handledGlyphs:
                continue
            handledGlyphs.add(glyph)
            glyph.addObserver(self, "_glyphChanged", "Glyph.Changed")

    def _unsubscribeFromGlyphs(self):
        handledGlyphs = set()
        for glyph in self._glyphs:
            if glyph in handledGlyphs:
                continue
            handledGlyphs.add(glyph)
            glyph.removeObserver(self, "Glyph.Changed")

    # widget

    def _cellEdited(self, row, col):
        if row == 0 or col == 0:
            return
        item = self.item(row, col).text()
        # Glyphs that do not have outlines leave empty cells, can't convert
        # that to a scalar
        if not item:
            return
        item = int(item)
        # -1 because the first col contains descriptive text
        glyph = self._glyphs[col - 1]
        if row == 1:
            glyph.width = item
        elif row == 2:
            glyph.leftMargin = item
        elif row == 3:
            glyph.rightMargin = item
        elif row == 4:
            prevIndex = col - 2
            if prevIndex >= 0 and glyph.font is not None:
                prevGlyph = self._glyphs[prevIndex]
                kerning = glyph.font.kerning
                kerning.write(prevGlyph, glyph, item)
        # defcon callbacks do the update

    def _itemChanged(self, current, previous):
        if current is not None:
            cur = current.column()
        if previous is not None:
            prev = previous.column()
            if current is not None and cur == prev:
                return
        self.colorColumn(None if current is None else cur)
        if current is not None:
            sel = cur - 1
        else:
            sel = None
        self.selectedIndexChanged.emit(sel)

    # --------------
    # Public methods
    # --------------

    def glyphs(self):
        return self._glyphs

    def setGlyphs(self, glyphs):
        self._unsubscribeFromGlyphs()
        self._glyphs = glyphs
        self._subscribeToGlyphs(self._glyphs)
        self.updateCells(False)

    def kerningEnabled(self):
        return self._kerningEnabled

    def setKerningEnabled(self, value):
        if value == self._kerningEnabled:
            return
        self._kerningEnabled = value
        for i in range(self.columnCount()):
            item = self.item(4, i)
            self.closePersistentEditor(item)
        self.updateCells()

    def updateCells(self, keepColor=True):
        self.blockSignals(True)
        self.setEditTriggers(QAbstractItemView.NoEditTriggers)
        coloredColumn = self._coloredColumn
        self.fillGlyphs()
        if (
            keepColor
            and coloredColumn is not None
            and coloredColumn < self.columnCount()
        ):
            self.colorColumn(coloredColumn)
        self.setEditTriggers(QAbstractItemView.CurrentChanged)
        self.blockSignals(False)

    def colorColumn(self, column):
        emptyBrush = QBrush(Qt.NoBrush)
        selectionColor = QColor(235, 235, 235)
        for i in range(self.rowCount()):
            if self._coloredColumn is not None:
                item = self.item(i, self._coloredColumn)
                # cached column might be invalid if user input deleted it
                if item is not None:
                    item.setBackground(emptyBrush)
            if column is not None:
                self.item(i, column).setBackground(selectionColor)
        self._coloredColumn = column

    def setCurrentGlyph(self, glyphIndex):
        # disable the widget to avoid stealing focus
        self.setEnabled(False)
        self.blockSignals(True)
        if glyphIndex is not None:
            # so we can scroll to the item
            self.setCurrentCell(1, glyphIndex + 1)
        self.setCurrentItem(None)
        if glyphIndex is not None:
            self.colorColumn(glyphIndex + 1)
        else:
            self.colorColumn(glyphIndex)
        self.blockSignals(False)
        self.setEnabled(True)

    def fillGlyphs(self):
        def metricsTableItem(content, disableCell=False):
            if isinstance(content, float):
                content = round(content)
            if content is not None:
                content = str(content)
            item = MetricsTableItem(content)
            if disableCell:
                item.setFlags(Qt.NoItemFlags)
            elif content is None:
                item.setFlags(Qt.ItemIsEnabled)
            # TODO: should fields be centered? I find left-aligned more
            # natural to read, personally...
            # item.setTextAlignment(Qt.AlignCenter)
            return item

        def getKern(previousGlyph, glyph):
            if previousGlyph is None:
                return ""
            font = glyph.font
            if font is None:
                return 0
            kerning = font.kerning
            if kerning is None:
                return 0
            return kerning.find(previousGlyph, glyph)

        self._coloredColumn = None
        self.setColumnCount(len(self._glyphs) + 1)
        prevGlyph = None
        for index, glyph in enumerate(self._glyphs):
            # TODO: see about allowing glyph name edit here
            self.setItem(0, index + 1, metricsTableItem(glyph.name, True))
            self.setItem(1, index + 1, metricsTableItem(glyph.width))
            self.setItem(2, index + 1, metricsTableItem(glyph.leftMargin))
            self.setItem(3, index + 1, metricsTableItem(glyph.rightMargin))
            kValue = getKern(prevGlyph, glyph)
            kDisabled = not (index and self._kerningEnabled)
            # TODO: grey out cells when disabled
            self.setItem(4, index + 1, metricsTableItem(kValue, kDisabled))
            self.setColumnWidth(index + 1, self._cellWidth)
            prevGlyph = glyph

    # ----------
    # Qt methods
    # ----------

    def closeEvent(self, event):
        super().closeEvent(event)
        if event.isAccepted():
            self._unsubscribeFromGlyphs()

    def sizeHint(self):
        # http://stackoverflow.com/a/7216486/2037879
        height = sum(self.rowHeight(k) for k in range(self.rowCount()))
        height += self.horizontalScrollBar().height()
        margins = self.contentsMargins()
        height += margins.top() + margins.bottom()
        return QSize(self.width(), height)

    def wheelEvent(self, event):
        # A mouse can only scroll along the y-axis. Use x-axis if we have one
        # (e.g. from touchpad), otherwise use y-axis.
        angleDelta = event.angleDelta().x() or event.angleDelta().y()
        cur = self.horizontalScrollBar().value()
        self.horizontalScrollBar().setValue(cur - angleDelta / 120)
        event.accept()



from PyQt5.QtCore import QSize, Qt
from PyQt5.QtWidgets import QApplication, QMessageBox, QStyle

from trufont.tools import platformSpecific


class MessageBox(QMessageBox):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setIcon(QMessageBox.Question)
        self.setWindowTitle(QApplication.applicationName())
        self.setWindowModality(Qt.WindowModal)
        if platformSpecific.showAppIconInDialog():
            iconSize = self.style().pixelMetric(
                QStyle.PM_MessageBoxIconSize, None, self
            )
            icon = self.windowIcon()
            size = icon.actualSize(QSize(iconSize, iconSize))
            self.setIconPixmap(icon.pixmap(size))


class ReloadMessageBox(MessageBox):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setText(self.tr("Do you want to reload the document?"))
        self.setInformativeText(self.tr("Your current changes will be lost."))
        self.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        self.setDefaultButton(QMessageBox.Yes)

    @classmethod
    def getReloadDocument(cls, parent, documentName=None):
        dialog = cls(parent)
        if documentName is not None:
            dialog.setText(
                dialog.tr("Do you want to reload “{}”?").format(documentName)
            )
        return dialog.exec_()


class CloseMessageBox(MessageBox):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setText(self.tr("Do you want to save your changes?"))
        self.setInformativeText(
            self.tr("Your changes will be lost if you don’t save them.")
        )
        self.setStandardButtons(
            QMessageBox.Save | QMessageBox.Discard | QMessageBox.Cancel
        )
        self.setDefaultButton(QMessageBox.Save)

    @classmethod
    def getCloseDocument(cls, parent, documentName=None):
        dialog = cls(parent)
        if documentName is not None:
            dialog.setText(
                dialog.tr(
                    "Do you want to save the changes you made " "to “{}”?"
                ).format(documentName)
            )
        return dialog.exec_()



import random

from hsluv import hsluv_to_rgb
from PyQt5.QtGui import QColor

# http://martin.ankerl.com/2009/12/09/how-to-create-random-colors-programmatically/  # noqa
golden_ratio_conjugate = 0.618033988749895


class ColorGenerator:
    """
    A static-state random color generator that uses the `HUSL color space`_.

    .. _`HUSL color space`: http://www.husl-colors.org
    """

    seed = random.random()
    # http://www.husl-colors.org/syntax/
    minS = random.uniform(30, 70)
    maxS = minS + 30
    minL = random.uniform(50, 70)
    maxL = minL + 20

    @classmethod
    def getColor(cls):
        cls.seed = (cls.seed + golden_ratio_conjugate) % 1
        hue = 360 * cls.seed
        sat = random.uniform(cls.minS, cls.maxS)
        lum = random.uniform(cls.minL, cls.maxL)
        return hsluv_to_rgb([hue, sat, lum])

    @classmethod
    def getQColor(cls):
        color = cls.getColor()
        return QColor.fromRgbF(*color)

    @classmethod
    def setSaturationFromRange(cls, lo, hi):
        rg = hi - lo
        cls.minS = random.uniform(lo, hi)
        cls.maxS = cls.minS + round(min(3 * rg / 4, 100 - hi))

    @classmethod
    def setLightnessFromRange(cls, lo, hi):
        rg = hi - lo
        cls.minL = random.uniform(lo, hi)
        cls.maxL = cls.minL + round(min(3 * rg / 4, 100 - hi))



from PyQt5.QtCore import QRectF, Qt
from PyQt5.QtGui import QPainter, QPainterPath, QPalette
from PyQt5.QtWidgets import QApplication, QRubberBand, QStyle, QStyleOptionRubberBand

from trufont.drawingTools.baseTool import BaseTool
from trufont.tools import platformSpecific

# Draw icon
_path = QPainterPath()
_path.setFillRule(Qt.WindingFill)
_path.addRect(6, 14, 10, 10)
_path.addEllipse(12, 6, 12, 12)


class ShapesTool(BaseTool):
    icon = _path
    name = QApplication.translate("ShapesTool", "Shapes")
    shortcut = "S"

    def __init__(self, parent=None):
        super().__init__(parent)
        self._startPoint = None
        self._endPoint = None
        self._rubberBandRect = None

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        self._glyph.beginUndoGroup()
        self._startPoint = event.localPos()

    def mouseMoveEvent(self, event):
        super().mouseMoveEvent(event)
        widget = self.parent()
        self._endPoint = event.localPos()

        # Get the width of the shape
        width = abs(int(self._endPoint.x() - self._startPoint.x()))

        # If Shift key is pressed, equalize the width and height of the shape
        if event.modifiers() & Qt.ShiftModifier:
            if self._startPoint.y() >= self._endPoint.y():
                self._endPoint.setY(int(self._startPoint.y() - width))
            else:
                self._endPoint.setY(int(self._startPoint.y() + width))

        # Draw a temporary shape guide
        self._rubberBandRect = QRectF(self._startPoint, self._endPoint)
        widget.update()

    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)

        # Remove the temporary shape guide
        self._rubberBandRect = None

        # Create a new contour
        contour = self._glyph.instantiateContour()
        self._glyph.appendContour(contour)
        self._glyph.selected = False

        # Get points to construct the new shape contour
        endX, endY = int(self._endPoint.x()), int(self._endPoint.y())
        startX, startY = int(self._startPoint.x()), int(self._startPoint.y())

        # Draw ellipse if right mouse button was pressed
        if event.button() == Qt.RightButton:
            handlePos = 0.55
            midX = (endX + startX) / 2
            midY = (endY + startY) / 2
            halfWidthX = (endX - startX) / 2
            halfWidthY = (endY - startY) / 2
            handleXStart = midX - halfWidthX * handlePos
            handleXEnd = midX + halfWidthX * handlePos
            handleYStart = midY - halfWidthY * handlePos
            handleYEnd = midY + halfWidthY * handlePos

            contour.addPoint((midX, startY), segmentType="curve", smooth=True)
            contour.addPoint((handleXEnd, startY))
            contour.addPoint((endX, handleYStart))

            contour.addPoint((endX, midY), segmentType="curve", smooth=True)
            contour.addPoint((endX, handleYEnd))
            contour.addPoint((handleXEnd, endY))

            contour.addPoint((midX, endY), segmentType="curve", smooth=True)
            contour.addPoint((handleXStart, endY))
            contour.addPoint((startX, handleYEnd))

            contour.addPoint((startX, midY), segmentType="curve", smooth=True)
            contour.addPoint((startX, handleYStart))
            contour.addPoint((handleXStart, startY))
            contour[0].selected = True

        # Else, draw a rectangle
        else:
            contour.addPoint((startX, startY), "line")
            contour.addPoint((startX, endY), "line")
            contour.addPoint((endX, endY), "line")
            contour.addPoint((endX, startY), "line")
            contour[0].selected = True

        self._glyph.endUndoGroup()

    # Draw the graphics on screen
    def paint(self, painter, index):
        if self._rubberBandRect is None:
            return
        widget = self.parent()
        if index != widget.activeIndex():
            return
        rect = self._rubberBandRect
        if platformSpecific.useBuiltinRubberBand():
            widgetOrigin = widget.mapFromCanvas(rect.bottomLeft())
            widgetMove = widget.mapFromCanvas(rect.topRight())
            option = QStyleOptionRubberBand()
            option.initFrom(widget)
            option.opaque = False
            option.rect = QRectF(widgetOrigin, widgetMove).toRect()
            option.shape = QRubberBand.Rectangle
            painter.save()
            painter.setRenderHint(QPainter.Antialiasing, False)
            painter.resetTransform()
            widget.style().drawControl(QStyle.CE_RubberBand, option, painter, widget)
            painter.restore()
        else:
            highlight = widget.palette().color(QPalette.Active, QPalette.Highlight)
            painter.save()
            painter.setRenderHint(QPainter.Antialiasing, False)
            pen = painter.pen()
            pen.setColor(highlight.darker(120))
            pen.setWidth(0)
            painter.setPen(pen)
            highlight.setAlphaF(0.35)
            painter.setBrush(highlight)
            painter.drawRect(rect)
            painter.restore()



"""
UI-constrained point management methods.
"""
import itertools

from PyQt5.QtCore import QLineF, QPointF

from trufont.tools import bezierMath
from trufont.tools.UIMove_ng import UIMove


def _getOffCurveSiblingPoints(contour, point):
    index = contour.index(point)
    pts = []
    # edge-cases: open contour boundaries and 3 offCurves+ qCurve segment
    shouldMoveAnyway = False
    onlyOffCurves = True
    for d in (-1, 1):
        sibling = contour.getPoint(index + d)
        if sibling.segmentType is not None:
            onlyOffCurves = False
        if sibling.selected:
            continue
        if sibling.segmentType is not None:
            sSibling = contour.getPoint(index + 2 * d)
            curPts = (sibling, sSibling)
            if contour.open and any(pt == contour[0] for pt in curPts):
                shouldMoveAnyway = True
                continue
            pts.append(curPts)
    if onlyOffCurves:
        shouldMoveAnyway = True
    return pts, shouldMoveAnyway


def maybeProjectUISmoothPointOffcurve(contour, index, delta=None):
    onCurve = contour[index]
    if not onCurve.smooth:
        return
    if contour.open and index in (0, len(contour) - 1):
        return
    offCurve, otherPoint = None, None
    for d in (-1, 1):
        pt = contour.getPoint(index + d)
        if pt.segmentType is None:
            if offCurve is not None:
                return
            offCurve = pt
        else:
            if otherPoint is not None:
                return
            if pt.selected and onCurve.selected:
                return
            px, py = pt.x, pt.y
            if d == 1 and delta is not None:
                # this point hasn't been moved yet. add the delta
                dx, dy = delta
                px += dx
                py += dy
            otherPoint = (px, py)
    if None not in (offCurve, otherPoint):
        px, py = otherPoint
        rotateUIPointAroundRefLine(px, py, onCurve.x, onCurve.y, offCurve)


def rotateUIPointAroundRefLine(x1, y1, x2, y2, pt):
    """
    Given three points p1, p2, pt this rotates pt around p2 such that p1,p2 and
    p1,pt are collinear.
    """
    line = QLineF(pt.x, pt.y, x2, y2)
    p2p_l = line.length()
    line.setP1(QPointF(x1, y1))
    p1p2_l = line.length()
    if not p1p2_l:
        return
    line.setLength(p1p2_l + p2p_l)
    pt.x = line.x2()
    pt.y = line.y2()


def moveUIPoint(contour, point, delta):
    if point.segmentType is None:
        # point is an offCurve. Get its sibling onCurve and the other
        # offCurve.
        siblings, shouldMoveAnyway = _getOffCurveSiblingPoints(contour, point)
        # if an onCurve is selected, the offCurve will move along with it
        if not (siblings or shouldMoveAnyway):
            return
        point.move(delta)
        for onCurve, otherPoint in siblings:
            if not onCurve.smooth or otherPoint.selected:
                continue
            # if the onCurve is smooth, we need to either...
            if otherPoint.segmentType is None:
                # keep the other offCurve inline
                rotateUIPointAroundRefLine(
                    point.x, point.y, onCurve.x, onCurve.y, otherPoint
                )
            else:
                # keep point in tangency with onCurve -> otherPoint segment,
                # i.e. do an orthogonal projection
                point.x, point.y, _ = bezierMath.lineProjection(
                    onCurve.x,
                    onCurve.y,
                    otherPoint.x,
                    otherPoint.y,
                    point.x,
                    point.y,
                    False,
                )
    else:
        # point is an onCurve. Move its offCurves along with it.
        index = contour.index(point)
        point.move(delta)
        for d in (-1, 1):
            # edge-case: contour open, trailing offCurve and moving first
            # onCurve in contour
            if contour.open and index == 0 and d == -1:
                continue
            pt = contour.getPoint(index + d)
            if pt.segmentType is None:
                # avoid double move for qCurve with single offCurve
                if d > 0:
                    otherPt = contour.getPoint(index + 2 * d)
                    if (
                        otherPt.segmentType is not None
                        and otherPt.segmentType != "move"
                        and otherPt.selected
                    ):
                        continue
                pt.move(delta)
                maybeProjectUISmoothPointOffcurve(contour, index, delta)
    contour.dirty = True


def deleteUISelection(glyph):
    for anchor in glyph.anchors:
        anchor.selected = not anchor.selected
    for component in glyph.components:
        component.selected = not component.selected
    for contour in glyph:
        for point in contour:
            point.selected = not point.selected
        contour.postNotification("Contour.SelectionChanged")
    cutGlyph = glyph.getRepresentation("TruFont.FilterSelection")
    glyph.beginUndoGroup()
    glyph.clear()
    pen = glyph.getPointPen()
    cutGlyph.drawPoints(pen)
    # HACK: defcon won't let us transfer anchors in bulk otherwise
    for anchor in cutGlyph.anchors:
        anchor._glyph = None
    glyph.anchors = cutGlyph.anchors
    glyph.endUndoGroup()


def removeUISelection(contour, preserveShape=True):
    segments = contour.segments
    # the last segments contains the first point, make sure to process it last
    # so as to not offset indexes
    toFirstPoint = segments[-1]
    toIter = list(enumerate(segments))
    toIter.insert(0, toIter.pop())
    # moonwalk through segments
    for index, segment in reversed(toIter):
        if segment == toFirstPoint:
            index = len(segments) - 1
        onCurve = segment[-1]
        # if the onCurve is selected, wipe it
        if onCurve.selected:
            # remove the contour if we have exhausted segments
            if len(segments) < 2:
                glyph = contour.glyph
                glyph.removeContour(contour)
                return
            # using preserveShape at the edge of an open contour will traceback
            # TODO: preserveShape should be implemented for qcurve
            if onCurve.segmentType in ("line", "qcurve"):
                preserveShape = False
            if preserveShape and contour.open:
                if index in (0, len(segments) - 1):
                    preserveShape = False
            contour.removeSegment(index, preserveShape)
            # remove segment so we can keep track of how many remain
            del segments[index]
        elif len(segment) == 2:
            # move with trailing offCurve
            offCurve = segment[0]
            if offCurve.selected:
                assert offCurve.segmentType is None
                contour.removePoint(offCurve)
        elif len(segment) == 3:
            # if offCurve selected, wipe them
            for i in (0, 1):
                if segment[i].selected:
                    onCurve = segment[2]
                    otherOnCurve = contour.getPoint(contour.index(onCurve) - 3)
                    for i in range(2):
                        contour.removePoint(segment[i])
                    onCurve.segmentType = "line"
                    onCurve.smooth = otherOnCurve.smooth = False
                    break


def UIGlyphGuidelines(glyph):
    guidelines = glyph.guidelines
    font = glyph.font
    if font is not None:
        guidelines = itertools.chain(guidelines, font.guidelines)
    return guidelines


def moveUIGlyphElements(glyph, dx, dy, **kwargs):
    delta = (dx, dy)
    glyph.beginUndoGroup()
    for anchor in glyph.anchors:
        if anchor.selected:
            anchor.move(delta)
    for contour in glyph:
        UIMove(contour, delta, **kwargs)
    for component in glyph.components:
        if component.selected:
            component.move(delta)
    for guideline in UIGlyphGuidelines(glyph):
        if guideline.selected:
            guideline.x += dx
            guideline.y += dy
    image = glyph.image
    if image.selected:
        image.move(delta)
    glyph.endUndoGroup()


def removeUIGlyphElements(glyph, preserveShape):
    glyph.beginUndoGroup()
    for anchor in glyph.anchors:
        if anchor.selected:
            glyph.removeAnchor(anchor)
    for contour in reversed(glyph):
        removeUISelection(contour, preserveShape)
    for component in glyph.components:
        if component.selected:
            glyph.removeComponent(component)
    for guideline in UIGlyphGuidelines(glyph):
        if guideline.selected:
            parent = guideline.getParent()
            parent.removeGuideline(guideline)
    if glyph.image.selected:
        glyph.image = None
    glyph.endUndoGroup()


def unselectUIGlyphElements(glyph):
    for anchor in glyph.anchors:
        anchor.selected = False
    for component in glyph.components:
        component.selected = False
    glyph.selected = False
    for guideline in UIGlyphGuidelines(glyph):
        guideline.selected = False
    glyph.image.selected = False



"""
The *platformSpecific* submodule
--------------------------------

While most of the time Qt abstracts any platform difference transparently,
there are times where it explicitly chooses not to (for instance, the Enter key
on Windows corresponds to the Return key on OSX) so to leave control to the
user.

All such occurrences are stored in the *platformSpecific* submodule to make
such code obvious and self-contained.

Fear not, these occurrences are rather anecdotic as you may tell from the size
of this file.
"""

import sys

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QFont, QFontDatabase

# ------
# Colors
# ------


def colorOpacityMultiplier():
    """
    Returns a transparency multiplier to have consistent system colors on all
    platforms. Assumes Windows = 1.
    """
    if sys.platform == "darwin":
        return 1.4
    return 1


# -----
# Fonts
# -----


def fixedFont():
    """
    Returns a default fixed-pitch QFont_ for each supported platform.

    Returns "Consolas" instead of the default "Courier New" on Windows.

    TODO: test more

    .. _QFont: http://doc.qt.io/qt-5/qfont.html
    """
    font = QFontDatabase.systemFont(QFontDatabase.FixedFont)
    if sys.platform == "win32":
        # pick Consolas instead of Courier New
        font.setFamily("Consolas")
    elif sys.platform == "darwin":
        # pick Menlo instead of Monaco
        font.setFamily("Menlo")
        font.setPointSize(11)
    return font


def otherUIFont():
    """
    Returns an auxiliary UI font.
    """
    font = QFont()
    pointSize = 9
    if sys.platform == "win32":
        font.setFamily("Segoe UI")
    elif sys.platform == "darwin":
        try:
            platform
        except NameError:
            import platform
        if platform.mac_ver()[0].startswith("10.10"):
            font.setFamily("Lucida Grande")
        pointSize = 12
    elif sys.platform.startswith("linux"):
        font.setFamily("Luxi Sans")
    font.setPointSize(pointSize)
    return font


# ----
# Keys
# ----


def scaleModifier():
    if sys.platform == "darwin":
        return Qt.AltModifier
    return Qt.ControlModifier



"""
The *glyphLineView* submodule
-----------------------------

The *glyphLineView* submodule provides widgets that render a list of Glyph_ or
:class:`GlyphRecord` along a line, with various display parameters.

.. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
"""

from defcon import Glyph
from PyQt5.QtCore import QRectF, QSize, Qt, pyqtSignal
from PyQt5.QtGui import QColor, QPainter, QPalette
from PyQt5.QtWidgets import QScrollArea, QSizePolicy, QWidget

from defconQt.controls.glyphCellView import cellGridColor
from defconQt.tools import drawing, platformSpecific


class GlyphLineWidget(QWidget):
    """
    The :class:`GlyphLineWidget` widget displays a list of :class:`GlyphRecord`
    objects on one or multiple line(s).

    Note that this widget does not perform text layout by itself, but the
    :class:`GlyphRecord` data structure can hold positioning information from
    a text shaping engine.

    # TODO: drag and drop
    """

    glyphActivated = pyqtSignal(Glyph)
    pointSizeModified = pyqtSignal(int)
    selectionModified = pyqtSignal(object)  # Note: object because it can be None

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAttribute(Qt.WA_OpaquePaintEvent)
        self.setFocusPolicy(Qt.ClickFocus)
        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        self._showLayers = False
        self._layerDrawingAttributes = {}
        self._fallbackDrawingAttributes = dict(
            showGlyphFill=True,
            showGlyphStroke=False,
            showGlyphOnCurvePoints=False,
            showGlyphStartPoints=False,
            showGlyphOffCurvePoints=False,
            showGlyphPointCoordinates=False,
            showGlyphAnchors=False,
            showGlyphImage=False,
            showGlyphMargins=False,
            showFontVerticalMetrics=False,
            showFontPostscriptBlues=False,
            showFontPostscriptFamilyBlues=False,
        )

        self._pointSize = 200
        self._scale = 1.0
        self._inverseScale = 0.1
        self._upm = 1000
        self._descender = -250
        self._buffer = 15

        self._drawMetrics = False
        self._verticalFlip = False
        self._lineHeight = 1.1
        self._rightToLeft = False
        self._wrapLines = False

        self._backgroundColor = Qt.white
        self._glyphColor = Qt.black
        self._glyphSelectionColor = None
        self._metricsColor = cellGridColor
        self._notdefBackgroundColor = QColor(255, 204, 204)

        self._glyphRecords = []
        self._glyphRecordsRects = {}
        self._selected = None

    # ------------
    # External API
    # ------------

    def scrollArea(self):
        return self._scrollArea

    def setScrollArea(self, scrollArea):
        scrollArea.setWidget(self)
        self._scrollArea = scrollArea

    def glyphRecords(self):
        """
        Returns the list of :class:`GlyphRecord` in the widgets. This may be
        empty.
        """
        return self._glyphRecords

    def setGlyphRecords(self, glyphRecords):
        """
        Sets the glyphRecords displayed by the widget to *glyphs*.

        *glyphs* may be a list of :class:`GlyphRecord` or a list of Glyph_, in
        which case they will be wrapped into :class:`GlyphRecord` classes with
        the font’s kerning applied if the *applyKerning* attribute is true.

        .. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
        """
        self._glyphRecordsRects = {}
        if self._selected is not None:
            selectedGlyph = self._glyphRecords[self._selected].glyph
        else:
            selectedGlyph = None
        # set the records into the view
        self._glyphRecords = glyphRecords
        # font metrics
        upms = []
        descenders = []
        for glyphRecord in self._glyphRecords:
            glyph = glyphRecord.glyph
            font = glyph.font
            if font is not None:
                upm = font.info.unitsPerEm
                if upm is not None:
                    upms.append(upm)
                descender = font.info.descender
                if descender is not None:
                    descenders.append(descender)
        if upms:
            self._upm = max(upms)
        if descenders:
            self._descender = min(descenders)
        # selection
        selectedIndex = None
        if self._selected is not None:
            if self._selected < len(self._glyphRecords):
                glyph = self._glyphRecords[self._selected].glyph
                if glyph == selectedGlyph:
                    selectedIndex = self._selected
        self.setSelected(selectedIndex)
        self._calcScale()
        self.setShowLayers(self._showLayers)
        self.adjustSize()
        self.update()

    def selected(self):
        """
        Returns the currently selected glyph’s index, or None if none are
        selected.
        """
        return self._selected

    def setSelected(self, value):
        """
        Sets *value*-th glyph as the selected glyph, or none if *value* is
        None.

        *value* should be less than the number of glyphRecords present in the
        widget.
        """
        self._selected = value
        if self._selected is not None and self._glyphRecordsRects is not None:
            scrollArea = self._scrollArea
            if scrollArea is not None:
                rect = None
                for r, indexRecord in self._glyphRecordsRects.items():
                    if indexRecord == self._selected:
                        rect = QRectF(*r)
                        break
                if rect is not None:
                    center = rect.center()
                    scrollArea.ensureVisible(
                        center.x(), center.y(), 0.6 * rect.width(), 0.6 * rect.height()
                    )
        self.update()

    def pointSize(self):
        """
        Returns this widget’s *pointSize*.
        """
        return self._pointSize

    def setPointSize(self, pointSize):
        """
        Sets the point size to *pointSize*. Must be greater than zero.

        The initial value is 150.
        """
        self._pointSize = int(pointSize)
        self._calcScale()
        self.adjustSize()
        self.update()

    def rightToLeft(self):
        """
        Returns this widget’s layout direction. Right-to-left if true.
        """
        return self._rightToLeft

    def setRightToLeft(self, value):
        """
        Sets the widget’s layout direction to right-to-left if *value* is true.

        The default is false.
        """
        self._rightToLeft = value
        self.adjustSize()
        self.update()

    def glyphColor(self):
        """
        Returns the color with which glyphs are rendered.
        """
        return self._glyphColor

    def setGlyphColor(self, color):
        """
        Sets the QColor_ *color* with which glyphs are rendered.

        The default is Qt.black.

        .. _QColor: http://doc.qt.io/qt-5/qcolor.html
        """
        self._glyphColor = color
        self.update()

    def metricsColor(self):
        return self._metricsColor

    def setMetricsColor(self, color):
        self._metricsColor = color
        self.update()

    def backgroundColor(self):
        """
        Returns the color with which the widget’s background is paint.
        """
        return self._backgroundColor

    def setBackgroundColor(self, color):
        """
        Sets the QColor_ *backgroundColor* with which the widget’s background
        is paint.

        The default is Qt.white.

        .. _QColor: http://doc.qt.io/qt-5/qcolor.html
        """
        self._backgroundColor = color
        self.update()

    def notdefBackgroundColor(self):
        """
        Returns the color with which .notdef’s background is paint.
        """
        return self._notdefBackgroundColor

    def setNotdefBackgroundColor(self, color):
        """
        Sets the QColor_ *notdefBackgroundColor* with which .notdef’s
        background is paint.

        The default is QColor(255, 204, 204).

        .. _QColor: http://doc.qt.io/qt-5/qcolor.html
        """
        self._notdefBackgroundColor = color
        self.update()

    def showLayers(self):
        """
        Returns whether an overlay of each rendered glyph’s layers if
        displayed.
        """
        return self._showLayers

    def setShowLayers(self, value):
        """
        Displays an overlay of each rendered glyph’s layers if *value* is true.

        The default is false.
        """
        self._layerDrawingAttributes = {}
        self._showLayers = value
        if value:
            for record in self._glyphRecords:
                glyph = record.glyph
                layerSet = glyph.layerSet
                if layerSet is not None:
                    for layerName in layerSet.layerOrder:
                        self._layerDrawingAttributes[layerName] = dict(
                            showGlyphFill=True
                        )
        self.update()

    def drawMetrics(self):
        return self._drawMetrics

    def setDrawMetrics(self, value):
        self._drawMetrics = value
        self.update()

    def verticalFlip(self):
        """
        Returns whether glyphs are displayed vertically flipped.
        """
        return self._verticalFlip

    def setVerticalFlip(self, value):
        """
        Displays glyphs vertically flipped if *value* is true.

        The default is false.
        """
        self._verticalFlip = value
        self.update()

    def lineHeight(self):
        """
        Returns this widget’s line-height factor.
        """
        return self._lineHeight

    def setLineHeight(self, scale):
        """
        Sets the line-height factor to *scale*. 1.0 means no extra vertical
        space is added around each line of text.

        The default value is 1.1.
        """
        self._lineHeight = scale
        self.adjustSize()
        self.update()

    def wrapLines(self):
        """
        Returns whether glyph runs wrap at the viewport’s *width* boundary.
        """
        return self._wrapLines

    def setWrapLines(self, value):
        """
        Sets the widget to wrap glyphs at the viewport’s *width* boundary if
        *value* is true.

        The default is false.
        """
        self._wrapLines = value
        scrollArea = self._scrollArea
        if scrollArea is not None:
            if value:
                scrollArea.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
                scrollArea.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
            else:
                scrollArea.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
                scrollArea.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.adjustSize()
        self.update()

    def drawingAttribute(self, attr, layerName):
        """
        Returns the suitable drawing attribute for string *attr* and
        *layerName*.

        If *layerName* is None, this returns a fallback attribute.
        """
        if layerName is None:
            return self._fallbackDrawingAttributes.get(attr)
        d = self._layerDrawingAttributes.get(layerName, {})
        return d.get(attr)

    def setDrawingAttribute(self, attr, value, layerName):
        """
        Sets attribute string *attr* to *value*. If *layerName* is not None,
        this value is specified for that given layer.

        The default values are:

        - *showGlyphFill=True*
        - *showGlyphStroke=False*
        - *showGlyphOnCurvePoints=False*
        - *showGlyphStartPoints=False*
        - *showGlyphOffCurvePoints=False*
        - *showGlyphPointCoordinates=False*
        - *showGlyphAnchors=False*
        - *showGlyphImage=False*
        - *showGlyphMargins=False*
        - *showFontVerticalMetrics=False*
        - *showFontPostscriptBlues=False*
        - *showFontPostscriptFamilyBlues=False*

        """
        if layerName is None:
            self._fallbackDrawingAttributes[attr] = value
        else:
            if layerName not in self._layerDrawingAttributes:
                self._layerDrawingAttributes[layerName] = {}
            self._layerDrawingAttributes[layerName][attr] = value
        self.update()

    # ------
    # Sizing
    # ------

    def _calcScale(self):
        scale = self._pointSize / self._upm
        if scale < 0.01:
            scale = 0.01
        self._scale = scale
        self._inverseScale = 1.0 / self._scale
        self._pointSize = self._upm * self._scale

    def sizeHint(self):
        width = height = self._buffer * 2
        if not self._glyphRecords:
            width += 600
            height += 200
        elif not self._wrapLines:
            for glyphRecord in self._glyphRecords:
                width += (
                    glyphRecord.advanceWidth
                    + glyphRecord.xPlacement
                    + glyphRecord.xAdvance
                ) * self._scale
            height += self._pointSize * self._lineHeight
        else:
            curWidth = self._buffer * 2
            scrollArea = self._scrollArea
            if scrollArea:
                visibleWidth = scrollArea.viewport().width()
            else:
                visibleWidth = self.width()
            lines = 1
            for glyphRecord in self._glyphRecords:
                glyph = glyphRecord.glyph
                gWidth = (
                    glyphRecord.advanceWidth
                    + glyphRecord.xPlacement
                    + glyphRecord.xAdvance
                ) * self._scale
                if curWidth + gWidth > visibleWidth or glyph.unicode == 2029:
                    curWidth = self._buffer * 2 + gWidth
                    lines += 1
                else:
                    curWidth += gWidth
                width = max(curWidth, width)
            height += lines * self._pointSize * self._lineHeight
        return QSize(width, height)

    # ------------
    # Input events
    # ------------

    def mouseDoubleClickEvent(self, event):
        if event.button() == Qt.LeftButton:
            if self._selected is not None:
                glyphRecord = self._glyphRecords[self._selected]
                self.glyphActivated.emit(glyphRecord.glyph)
        else:
            super().mouseDoubleClickEvent(event)

    def wheelEvent(self, event):
        if event.modifiers() & platformSpecific.scaleModifier():
            scale = pow(1.2, event.angleDelta().y() / 120.0)
            pointSize = round(self._pointSize * scale)
            if pointSize <= 0:
                return
            self.setPointSize(pointSize)
            self.pointSizeModified.emit(pointSize)
            event.accept()
        else:
            super().wheelEvent(event)

    # --------
    # Painting
    # --------

    def drawLineBackground(self, painter, rect):
        if self._drawMetrics:
            descender = self._descender
            ascender = self._upm + descender
            x, _, width, _ = rect
            painter.save()
            pen = painter.pen()
            pen.setColor(self._metricsColor)
            painter.setPen(pen)
            drawing.drawLine(painter, x, ascender, width, ascender)
            drawing.drawLine(painter, x, 0, width, 0)
            drawing.drawLine(painter, x, descender, width, descender)
            painter.restore()

    def drawGlyph(self, painter, glyph, rect, selected=False):
        # gather the layers
        layerSet = glyph.layerSet
        if layerSet is None or not self._showLayers:
            layers = [(glyph, None)]
        else:
            glyphName = glyph.name
            layers = []
            for layerName in reversed(layerSet.layerOrder):
                layer = layerSet[layerName]
                if glyphName not in layer:
                    continue
                g = layer[glyphName]
                if g == glyph:
                    layerName = None
                layers.append((g, layerName))

        self.drawGlyphBackground(painter, glyph, rect, selected=selected)

        for g, layerName in layers:
            # draw the image
            if self.drawingAttribute("showGlyphImage", layerName):
                self.drawImage(painter, g, layerName, rect)
            # draw the blues
            if layerName is None and self.drawingAttribute(
                "showFontPostscriptBlues", None
            ):
                self.drawBlues(painter, g, layerName, rect)
            if layerName is None and self.drawingAttribute(
                "showFontPostscriptFamilyBlues", None
            ):
                self.drawFamilyBlues(painter, g, layerName, rect)
            # draw the margins
            if self.drawingAttribute("showGlyphMargins", layerName):
                self.drawMargins(painter, g, layerName, rect)
            # draw the vertical metrics
            if layerName is None and self.drawingAttribute(
                "showFontVerticalMetrics", None
            ):
                self.drawVerticalMetrics(painter, g, layerName, rect)
            # draw the glyph
            if self.drawingAttribute(
                "showGlyphFill", layerName
            ) or self.drawingAttribute("showGlyphStroke", layerName):
                self.drawFillAndStroke(painter, g, layerName, rect)
            if self.drawingAttribute(
                "showGlyphOnCurvePoints", layerName
            ) or self.drawingAttribute("showGlyphOffCurvePoints", layerName):
                self.drawPoints(painter, g, layerName, rect)
            if self.drawingAttribute("showGlyphAnchors", layerName):
                self.drawAnchors(painter, g, layerName, rect)

        self.drawGlyphForeground(painter, glyph, rect, selected=selected)

    def drawGlyphBackground(self, painter, glyph, rect, selected=False):
        if glyph.name == ".notdef":
            painter.fillRect(QRectF(*rect), self._notdefBackgroundColor)
        if selected:
            if self._glyphSelectionColor is not None:
                selectionColor = self._glyphSelectionColor
            else:
                palette = self.palette()
                active = palette.currentColorGroup() != QPalette.Inactive
                opacityMultiplier = platformSpecific.colorOpacityMultiplier()
                selectionColor = palette.color(QPalette.Highlight)
                selectionColor.setAlphaF(0.2 * opacityMultiplier if active else 0.9)
            xMin, yMin, width, height = rect
            painter.save()
            if self._drawMetrics:
                pen = painter.pen()
                pen.setStyle(Qt.DotLine)
                pen.setColor(self._metricsColor)
                painter.setPen(pen)
                drawing.drawLine(painter, xMin, yMin, xMin, yMin + height)
                drawing.drawLine(
                    painter, xMin + width, yMin, xMin + width, yMin + height
                )
            painter.fillRect(xMin, yMin, width, -26, selectionColor)
            painter.restore()

    def drawImage(self, painter, glyph, layerName, rect):
        drawing.drawGlyphImage(
            painter, glyph, self._inverseScale, backgroundColor=self._backgroundColor
        )

    def drawBlues(self, painter, glyph, layerName, rect):
        drawing.drawFontPostscriptBlues(
            painter, glyph, self._inverseScale, backgroundColor=self._backgroundColor
        )

    def drawFamilyBlues(self, painter, glyph, layerName, rect):
        drawing.drawFontPostscriptFamilyBlues(
            painter, glyph, self._inverseScale, backgroundColor=self._backgroundColor
        )

    def drawVerticalMetrics(self, painter, glyph, layerName, rect):
        drawText = (
            self.drawingAttribute("showFontVerticalMetricsTitles", layerName)
            and self._pointSize > 150
        )
        drawing.drawFontVerticalMetrics(
            painter,
            glyph,
            self._inverseScale,
            drawText=drawText,
            backgroundColor=self._backgroundColor,
        )

    def drawMargins(self, painter, glyph, layerName, rect):
        drawing.drawGlyphMargins(
            painter, glyph, self._inverseScale, backgroundColor=self._backgroundColor
        )

    def drawFillAndStroke(self, painter, glyph, layerName, rect):
        showFill = self.drawingAttribute("showGlyphFill", layerName)
        showStroke = self.drawingAttribute("showGlyphStroke", layerName)
        fillColor = None
        if not self._showLayers:
            fillColor = self._glyphColor
        drawing.drawGlyphFillAndStroke(
            painter,
            glyph,
            self._inverseScale,
            drawFill=showFill,
            drawStroke=showStroke,
            contourFillColor=fillColor,
            componentFillColor=fillColor,
        )

    def drawPoints(self, painter, glyph, layerName, rect):
        drawStartPoints = (
            self.drawingAttribute("showGlyphStartPoints", layerName)
            and self._pointSize > 175
        )
        drawOnCurves = (
            self.drawingAttribute("showGlyphOnCurvePoints", layerName)
            and self._pointSize > 175
        )
        drawOffCurves = (
            self.drawingAttribute("showGlyphOffCurvePoints", layerName)
            and self._pointSize > 175
        )
        drawCoordinates = (
            self.drawingAttribute("showGlyphPointCoordinates", layerName)
            and self._pointSize > 250
        )
        drawing.drawGlyphPoints(
            painter,
            glyph,
            self._inverseScale,
            drawStartPoints=drawStartPoints,
            drawOnCurves=drawOnCurves,
            drawOffCurves=drawOffCurves,
            drawCoordinates=drawCoordinates,
            backgroundColor=self._backgroundColor,
        )

    def drawAnchors(self, painter, glyph, layerName, rect):
        drawText = self._pointSize > 50
        drawing.drawGlyphAnchors(
            painter,
            glyph,
            self._inverseScale,
            drawText=drawText,
            backgroundColor=self._backgroundColor,
        )

    def drawGlyphForeground(self, painter, glyph, rect, selected=False):
        pass

    # ------
    # Events
    # ------

    def keyPressEvent(self, event):
        key = event.key()
        if key == Qt.Key_Return:
            index = self._selected
            if index is not None:
                glyph = self._glyphRecords[index].glyph
                self.glyphActivated.emit(glyph)
        else:
            super().keyPressEvent(event)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            selected = None
            for rect, recordIndex in self._glyphRecordsRects.items():
                if QRectF(*rect).contains(event.localPos()):
                    selected = recordIndex
            if self._selected == selected:
                return
            self._selected = selected
            self.selectionModified.emit(self._selected)
            self.update()
        else:
            super().mousePressEvent(event)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        # TODO: we could possibly redo rects only on adjustSize()
        if self._rightToLeft:
            self.paintRightToLeft(painter, event.rect())
        else:
            self.paintLeftToRight(painter, event.rect())

    def paintLeftToRight(self, painter, rect):
        self._glyphRecordsRects = {}
        # draw the background
        painter.fillRect(rect, self._backgroundColor)
        if not self._glyphRecords:
            return
        # create some reusable values
        scale = self._scale
        upm = self._upm
        descender = self._descender
        ascender = upm + descender
        lineHeight = _roundScale(self._lineHeight, self._pointSize)
        lineHeightOffset = (lineHeight - 1) * upm
        # offset for the buffer
        painter.save()
        painter.translate(self._buffer, self._buffer)
        # offset for the descender
        baselineShift = 0.5 * lineHeightOffset
        if self._verticalFlip:
            baselineShift -= descender
            yDirection = 1
        else:
            baselineShift += ascender
            yDirection = -1
        baselineShift = _roundScale(baselineShift, scale)
        painter.scale(scale, scale)
        painter.translate(0, baselineShift)
        # flip
        painter.scale(1, yDirection)
        # draw metrics lines
        lineRect = (
            -self._buffer * self._inverseScale,
            descender,
            self.width() * self._inverseScale,
            upm,
        )
        self.drawLineBackground(painter, lineRect)
        # draw the records
        left = self._buffer
        top = self._buffer + 0.5 * lineHeightOffset * scale
        height = upm * scale
        for recordIndex, glyphRecord in enumerate(self._glyphRecords):
            glyph = glyphRecord.glyph
            w = glyphRecord.advanceWidth
            h = glyphRecord.advanceHeight
            xP = glyphRecord.xPlacement
            yP = glyphRecord.yPlacement
            xA = glyphRecord.xAdvance
            yA = glyphRecord.yAdvance
            # possibly go to the next line
            if self._wrapLines:
                incomingWidth = left + (w + xP + xA) * scale + self._buffer
                if incomingWidth > self.width() or glyph.unicode == 2029:
                    top += upm * lineHeight * scale
                    painter.translate(
                        (self._buffer - left) * self._inverseScale,
                        yDirection * upm * lineHeight,
                    )
                    left = self._buffer
                    self.drawLineBackground(painter, lineRect)
            # handle offsets from the record
            top -= yP * scale
            glyphHeight = height + ((h + yA) * scale)
            glyphLeft = left + (xP * scale)
            glyphWidth = (w + xA) * scale
            # store the glyph rect for the alternate menu
            rect = (glyphLeft, top, glyphWidth, glyphHeight)
            self._glyphRecordsRects[rect] = recordIndex
            # handle placement
            if xP or yP:
                painter.translate(xP, yP)
            # draw the glyph
            rect = (-xP, descender - yP, w, upm)
            selected = self._selected == recordIndex
            self.drawGlyph(painter, glyph, rect, selected=selected)
            # shift for the next glyph
            painter.translate(w + xA - xP, h + yA - yP)
            left += glyphWidth
        painter.restore()

    def paintRightToLeft(self, painter, rect):
        self._glyphRecordsRects = {}
        # draw the background
        painter.fillRect(rect, self._backgroundColor)
        # create some reusable values
        scale = self._scale
        upm = self._upm
        descender = self._descender
        ascender = upm + descender
        lineHeight = _roundScale(self._lineHeight, self._pointSize)
        lineHeightOffset = (lineHeight - 1) * upm
        # offset for the buffer
        painter.save()
        painter.translate(self.width() - self._buffer, self._buffer)
        # offset for the descender
        baselineShift = 0.5 * lineHeightOffset
        if self._verticalFlip:
            baselineShift -= descender
            yDirection = 1
        else:
            baselineShift += ascender
            yDirection = -1
        baselineShift = _roundScale(baselineShift, scale)
        painter.scale(scale, scale)
        painter.translate(0, baselineShift)
        # flip
        painter.scale(1, yDirection)
        # draw the records
        left = self.width() - self._buffer
        top = self._buffer + 0.5 * lineHeightOffset * scale
        height = upm * scale
        previousXA = 0
        for recordIndex, glyphRecord in enumerate(self._glyphRecords):
            glyph = glyphRecord.glyph
            w = glyphRecord.advanceWidth
            h = glyphRecord.advanceHeight
            xP = glyphRecord.xPlacement
            yP = glyphRecord.yPlacement
            xA = glyphRecord.xAdvance
            yA = glyphRecord.yAdvance
            # possibly go to the next line
            if self._wrapLines:
                incomingLeft = left - (w + xP + xA) * scale - self._buffer
                if incomingLeft < 0 or glyph.unicode == 2029:
                    top += upm * lineHeight * scale
                    painter.translate(
                        (self.width() - self._buffer - left) * self._inverseScale,
                        yDirection * upm * lineHeight,
                    )
                    left = self.width() - self._buffer
            # handle offsets from the record
            top -= yP * scale
            glyphHeight = height + ((h + yA) * scale)
            glyphLeft = left + ((-w + xP - xA) * scale)
            glyphWidth = (w + xA) * scale
            # store the glyph rect for the alternate menu
            rect = (glyphLeft, top, glyphWidth, glyphHeight)
            self._glyphRecordsRects[rect] = recordIndex
            # handle the placement
            if xP:
                xP += previousXA
            painter.translate(-w - xA + xP, yP)
            # draw the glyph
            rect = (-xP, descender - yP, w, upm)
            selected = self._selected == recordIndex
            self.drawGlyph(painter, glyph, rect, selected=selected)
            # shift for the next glyph
            painter.translate(-xP, h + yA - yP)
            left -= (w + xP + xA) * scale
            previousXA = xA
        painter.restore()


def _roundScale(value, scale):
    return round(value * scale) / scale


class GlyphLineView(QScrollArea):
    """
    The :class:`GlyphLineView` widget is a QScrollArea_ that contains a
    :class:`GlyphLineWidget`.

    It reimplements :class:`GlyphLineWidget` public API and handles
    notification support from the font.
    Its :func:`setGlyphs` function can wrap a list of Glyph_ into
    :class:`GlyphRecord` for convenience, and applies kerning values from the
    font if *applyKerning* is set to true in the constructor.

    Here’s a basic text example that shows a :class:`GlyphLineView` in a
    :class:`BaseWindow`:

    >>> from defconQt.controls.glyphLineView import GlyphLineView
    >>> window = BaseWindow()
    >>> glyphs = [font[c] for c in "hallo welt"]
    >>> area = GlyphLineView(window)
    >>> area.setGlyphs(glyphs)
    >>>
    >>> layout = QVBoxLayout(window)
    >>> layout.addWidget(area)
    >>> window.show()

    TODO: add sample image

    .. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
    .. _QScrollArea: http://doc.qt.io/qt-5/qscrollarea.html
    """

    glyphLineWidgetClass = GlyphLineWidget

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setWidgetResizable(True)

        self._applyKerning = False
        self._glyphLineWidget = self.glyphLineWidgetClass(self)
        self._glyphLineWidget.setScrollArea(self)
        # reexport signals
        self.glyphActivated = self._glyphLineWidget.glyphActivated
        self.pointSizeModified = self._glyphLineWidget.pointSizeModified
        self.selectionModified = self._glyphLineWidget.selectionModified

    # -------------
    # Notifications
    # -------------

    def _subscribeToGlyphs(self, glyphRecords):
        handledGlyphs = set()
        handledFonts = set()
        for glyphRecord in glyphRecords:
            glyph = glyphRecord.glyph
            if glyph in handledGlyphs:
                continue
            handledGlyphs.add(glyph)
            glyph.addObserver(self, "_glyphChanged", "Glyph.Changed")
            font = glyph.font
            if font is None:
                continue
            if font in handledFonts:
                continue
            handledFonts.add(font)
            font.info.addObserver(self, "_fontChanged", "Info.Changed")
            if self._applyKerning:
                font.kerning.addObserver(self, "_kerningChanged", "Kerning.Changed")

    def _unsubscribeFromGlyphs(self):
        handledGlyphs = set()
        handledFonts = set()
        glyphRecords = self._glyphLineWidget.glyphRecords()
        for glyphRecord in glyphRecords:
            glyph = glyphRecord.glyph
            if glyph in handledGlyphs:
                continue
            handledGlyphs.add(glyph)
            glyph.removeObserver(self, "Glyph.Changed")
            font = glyph.font
            if font is None:
                continue
            if font in handledFonts:
                continue
            handledFonts.add(font)
            font.info.removeObserver(self, "Info.Changed")
            if self._applyKerning:
                font.kerning.removeObserver(self, "Kerning.Changed")

    def _glyphChanged(self, notification):
        # self._glyphLineWidget.update()
        # TODO: figure out whether to use advanceWidth at all or not
        glyphRecords = self._glyphLineWidget.glyphRecords()
        for glyphRecord in glyphRecords:
            glyph = glyphRecord.glyph
            glyphRecord.advanceWidth = glyph.width
        self._glyphLineWidget.setGlyphRecords(glyphRecords)

    def _kerningChanged(self, notification):
        glyphRecords = self._glyphLineWidget.glyphRecords()
        self._setKerningInGlyphRecords(glyphRecords)
        self._glyphLineWidget.update()

    def _fontChanged(self, notification):
        glyphRecords = self._glyphLineWidget.glyphRecords()
        self._glyphLineWidget.setGlyphRecords(glyphRecords)

    # ---------------
    # Kerning Support
    # ---------------

    def _lookupKerningValue(self, font, first, second):
        # TODO: this should be put into defcon
        kerning = font.kerning
        groups = font.groups
        # quickly check to see if the pair is in the kerning dictionary
        pair = (first, second)
        if pair in kerning:
            return kerning[pair]
        # get group names and make sure first and second are glyph names
        firstGroup = secondGroup = None
        if first.startswith("public.kern1"):
            firstGroup = first
            first = None
        else:
            for group, groupMembers in groups.items():
                if group.startswith("public.kern1"):
                    if first in groupMembers:
                        firstGroup = group
                        break
        if second.startswith("public.kern2"):
            secondGroup = second
            second = None
        else:
            for group, groupMembers in groups.items():
                if group.startswith("public.kern2"):
                    if second in groupMembers:
                        secondGroup = group
                        break
        # make an ordered list of pairs to look up
        pairs = [
            (first, second),
            (first, secondGroup),
            (firstGroup, second),
            (firstGroup, secondGroup),
        ]
        # look up the pairs and return any matches
        for pair in pairs:
            if pair in kerning:
                return kerning[pair]
        return 0

    def _setKerningInGlyphRecords(self, glyphRecords):
        previousGlyph = None
        previousFont = None
        for index, glyphRecord in enumerate(glyphRecords):
            glyph = glyphRecord.glyph
            font = glyph.font
            if (
                previousGlyph is not None
                and font is not None
                and (previousFont == font)
            ):
                kern = self._lookupKerningValue(font, previousGlyph.name, glyph.name)
                if kern is None:
                    kern = 0
                glyphRecords[index - 1].xAdvance = kern
            previousGlyph = glyph
            previousFont = font

    # ------------
    # External API
    # ------------

    def applyKerning(self):
        return self._applyKerning

    def setApplyKerning(self, value):
        self._applyKerning = value

    def glyphRecords(self):
        return self._glyphLineWidget.glyphRecords()

    def setGlyphRecords(self, glyphs):
        # unsubscribe from the old glyphs
        self._unsubscribeFromGlyphs()
        # test to see if glyph records are present
        needToWrap = False
        if glyphs:
            for attr in (
                "glyph",
                "xPlacement",
                "yPlacement",
                "xAdvance",
                "yAdvance",
                "alternates",
            ):
                if not hasattr(glyphs[0], attr):
                    needToWrap = True
                    break
        # wrap into glyph records if necessary
        if needToWrap:
            glyphRecords = []
            for glyph in glyphs:
                glyphRecord = GlyphRecord()
                glyphRecord.glyph = glyph
                glyphRecord.advanceWidth = glyph.width
                glyphRecords.append(glyphRecord)
            # apply kerning as needed
            if self._applyKerning:
                self._setKerningInGlyphRecords(glyphRecords)
        else:
            glyphRecords = glyphs
        # set the records into the view
        self._glyphLineWidget.setGlyphRecords(glyphRecords)
        # subscribe to the new glyphs
        self._subscribeToGlyphs(glyphRecords)

    def pointSize(self):
        return self._glyphLineWidget.pointSize()

    def setPointSize(self, pointSize):
        self._glyphLineWidget.setPointSize(pointSize)

    def rightToLeft(self):
        return self._glyphLineWidget.rightToLeft()

    def setRightToLeft(self, value):
        self._glyphLineWidget.setRightToLeft(value)

    def glyphColor(self):
        return self._glyphLineWidget.glyphColor()

    def setGlyphColor(self, color):
        self._glyphLineWidget.setGlyphColor(color)

    def backgroundColor(self):
        return self._glyphLineWidget.backgroundColor()

    def setBackgroundColor(self, color):
        self._glyphLineWidget.setBackgroundColor(color)

    def notdefBackgroundColor(self):
        return self._glyphLineWidget._notdefBackgroundColor

    def setNotdefBackgroundColor(self, color):
        self._glyphLineWidget.setNotdefBackgroundColor(color)

    def showLayers(self):
        return self._glyphLineWidget.showLayers()

    def setShowLayers(self, value):
        self._glyphLineWidget.setShowLayers(value)

    def drawMetrics(self):
        return self._glyphLineWidget.drawMetrics()

    def setDrawMetrics(self, value):
        self._glyphLineWidget.setDrawMetrics(value)

    def lineHeight(self):
        return self._glyphLineWidget.lineHeight()

    def setLineHeight(self, scale):
        self._glyphLineWidget.setLineHeight(scale)

    def verticalFlip(self):
        return self._glyphLineWidget.verticalFlip()

    def setVerticalFlip(self, value):
        self._glyphLineWidget.setVerticalFlip(value)

    def wrapLines(self):
        return self._glyphLineWidget.wrapLines()

    def setWrapLines(self, value):
        self._glyphLineWidget.setWrapLines(value)

    def selected(self):
        return self._glyphLineWidget.selected()

    def setSelected(self, value):
        self._glyphLineWidget.setSelected(value)

    def drawingAttribute(self, attr, layerName=None):
        return self._glyphLineWidget.drawingAttribute(attr, layerName)

    def setDrawingAttribute(self, attr, value, layerName=None):
        self._glyphLineWidget.setDrawingAttribute(attr, layerName, value)


# ------------------
# Basic Glyph Record
# ------------------


class GlyphRecord:
    """
    A :class:`GlyphRecord` is a glyph data structure that can accomodate
    information suitable to text layout, namely positioning and alternates.
    """

    __slots__ = [
        "glyph",
        "xPlacement",
        "yPlacement",
        "xAdvance",
        "yAdvance",
        "advanceWidth",
        "advanceHeight",
        "alternates",
    ]

    def __init__(self):
        self.glyph = None
        self.xPlacement = 0
        self.yPlacement = 0
        self.xAdvance = 0
        self.yAdvance = 0
        self.advanceWidth = 0
        self.advanceHeight = 0
        self.alternates = []



import os
import tokenize
import traceback
from keyword import kwlist

from PyQt5.QtCore import QDir, QSettings, QSize, Qt, QUrl, pyqtSignal
from PyQt5.QtGui import (
    QColor,
    QDesktopServices,
    QKeySequence,
    QTextCharFormat,
    QTextCursor,
)
from PyQt5.QtWidgets import (
    QApplication,
    QComboBox,
    QFileDialog,
    QFileSystemModel,
    QMainWindow,
    QMenu,
    QMessageBox,
    QPushButton,
    QShortcut,
    QSplitter,
    QStatusBar,
    QTreeView,
    QVBoxLayout,
    QWidget,
)

from defconQt.controls.baseCodeEditor import (
    BaseCodeEditor,
    BaseCodeHighlighter,
    GotoLineDialog,
)
from trufont.controls.clickLabel import ClickLabel
from trufont.controls.fileMessageBoxes import CloseMessageBox
from trufont.objects import settings
from trufont.objects.menu import Entries
from trufont.tools import platformSpecific
from trufont.windows.outputWindow import OutputEdit, OutputStream


class ScriptingWindow(QMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.editor = PythonEditor(self)
        self.fileChooser = FileChooser(self)
        self.fileChooser.fileOpened.connect(self.openFile)
        self.outputEdit = OutputEdit(self)

        splitter = QSplitter(self)
        self.vSplitter = QSplitter(Qt.Vertical, splitter)
        self.vSplitter.addWidget(self.editor)
        self.vSplitter.addWidget(self.outputEdit)
        self.vSplitter.setStretchFactor(0, 1)
        self.vSplitter.setStretchFactor(1, 0)
        splitter.addWidget(self.fileChooser)
        splitter.addWidget(self.vSplitter)
        splitter.setStretchFactor(0, 0)
        splitter.setStretchFactor(1, 1)
        statusBar = ScriptingStatusBar(self)
        self.setCentralWidget(splitter)
        self.setStatusBar(statusBar)
        self.newFile()
        self.editor.modificationChanged.connect(self.setWindowModified)
        statusBar.setPosition(self.editor.textCursor())
        self.editor.cursorPositionChanged.connect(
            lambda: statusBar.setPosition(self.sender().textCursor())
        )
        statusBar.setIndent(self.editor.indent())
        self.editor.indentChanged.connect(statusBar.setIndent)
        statusBar.indentModified.connect(self.editor.setIndent)
        statusBar.positionClicked.connect(self.gotoLine)
        statusBar.clearButtonClicked.connect(self.outputEdit.clear)
        statusBar.runButtonClicked.connect(self.runScript)
        gotoLineShortcut = QShortcut(QKeySequence("Ctrl+G"), self)
        gotoLineShortcut.activated.connect(self.gotoLine)

        self.readSettings()
        splitter.splitterMoved.connect(self.writeSettings)
        self.vSplitter.splitterMoved.connect(self.writeSettings)

    def readSettings(self):
        geometry = settings.scriptingWindowGeometry()
        if geometry:
            self.restoreGeometry(geometry)
        sizes = settings.scriptingWindowHSplitterSizes()
        if sizes:
            splitter = self.centralWidget()
            splitter.setSizes(sizes)
        sizes = settings.scriptingWindowVSplitterSizes()
        if sizes:
            self.vSplitter.setSizes(sizes)

    def writeSettings(self):
        settings.setScriptingWindowGeometry(self.saveGeometry())
        # splitters don't report a correct size until the window is visible
        if not self.isVisible():
            return
        splitter = self.centralWidget()
        settings.setScriptingWindowHSplitterSizes(splitter.sizes())
        settings.setScriptingWindowVSplitterSizes(self.vSplitter.sizes())

    def setupMenu(self, menuBar):
        fileMenu = menuBar.fetchMenu(Entries.File)
        fileMenu.fetchAction(Entries.File_New, self.newFile)
        fileMenu.fetchAction(Entries.File_Open, self.openFile)
        fileMenu.fetchAction(Entries.File_Save, self.saveFile)
        fileMenu.fetchAction(Entries.File_Save_As, self.saveFileAs)
        fileMenu.addSeparator()
        fileMenu.fetchAction(Entries.File_Close, self.close)

    @property
    def currentPath(self):
        return self._currentPath

    @currentPath.setter
    def currentPath(self, currentPath):
        self._currentPath = currentPath
        if self._currentPath is None:
            title = self.tr("Untitled")
        else:
            title = os.path.basename(self._currentPath)
        self.setWindowTitle(title)

    def newFile(self):
        if not self._maybeSaveBeforeExit():
            return
        self.editor.setPlainText(None)
        self.currentPath = None

    def openFile(self, path=None):
        if not self._maybeSaveBeforeExit():
            return
        if path is None:
            path = self._ioDialog(QFileDialog.AcceptOpen)
            if path is None:
                return
            self.fileChooser.setCurrentFolder(os.path.dirname(path))
        with tokenize.open(path) as inputFile:
            self.editor.setPlainText(inputFile.read())
        self.currentPath = path

    def saveFile(self):
        if self.currentPath is None:
            self.saveFileAs()
        else:
            self.editor.write(self.currentPath)

    def saveFileAs(self):
        path = self._ioDialog(QFileDialog.AcceptSave)
        if path is not None:
            self.currentPath = path
            self.saveFile()

    # TODO: why not use simple dialogs?
    def _ioDialog(self, mode):
        state = settings.scriptingFileDialogState()
        if mode == QFileDialog.AcceptOpen:
            title = self.tr("Open File")
        else:
            title = self.tr("Save File")
        dialog = QFileDialog(self, title, None, self.tr("Python file (*.py)"))
        if state:
            dialog.restoreState(state)
        dialog.setAcceptMode(mode)
        dialog.setDirectory(self.fileChooser.currentFolder())
        dialog.setFileMode(QFileDialog.ExistingFile)
        ok = dialog.exec_()
        settings.setScriptingFileDialogState(state)
        if ok:
            return dialog.selectedFiles()[0]
        return None

    def gotoLine(self):
        newLine, newColumn, ret = GotoLineDialog.getLineColumnNumber(self)
        if ret and newLine:
            self.editor.scrollToLine(newLine, newColumn)

    def runScript(self):
        app = QApplication.instance()
        global_vars = app.globals()
        script = self.editor.toPlainText()
        streams = []
        self.outputEdit.grayOutOldContent()
        for channel in ("stdout", "stderr"):
            stream = OutputStream(channel, self)
            stream.forward = True
            stream.messagePassed.connect(self.outputEdit.write)
            streams.append(stream)
        try:
            code = compile(script, "<string>", "exec")
            exec(code, global_vars)
        except Exception:
            traceback.print_exc()
        for stream in streams:
            stream.unregisterStream()
            stream.messagePassed.disconnect(self.outputEdit.write)

    # ----------
    # Qt methods
    # ----------

    def setWindowTitle(self, title):
        if platformSpecific.appNameInTitle():
            title += " – TruFont"
        super().setWindowTitle(f"[*]{title}")

    def sizeHint(self):
        return QSize(650, 700)

    def moveEvent(self, event):
        self.writeSettings()

    resizeEvent = moveEvent

    def closeEvent(self, event):
        ok = self._maybeSaveBeforeExit()
        if ok:
            event.accept()
        else:
            event.ignore()

    def _maybeSaveBeforeExit(self):
        if self.isWindowModified():
            currentFile = self.windowTitle()[3:]
            ret = CloseMessageBox.getCloseDocument(self, currentFile)
            if ret == QMessageBox.Save:
                self.saveFile()
                return True
            elif ret == QMessageBox.Discard:
                return True
            return False
        return True


class FileTreeView(QTreeView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.doubleClickCallback = None

    def _showInExplorer(self, path):
        if os.path.isfile(path):
            path = os.path.dirname(path)
        QDesktopServices.openUrl(QUrl.fromLocalFile(path))

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        if event.button() == Qt.RightButton:
            model = self.model()
            modelIndex = self.indexAt(event.pos())
            if modelIndex.isValid():
                path = model.filePath(modelIndex)
            else:
                path = model.rootPath()
            menu = QMenu(self)
            menu.addAction(
                self.tr("Open In Explorer"), lambda: self._showInExplorer(path)
            )
            menu.exec_(event.globalPos())

    def mouseDoubleClickEvent(self, event):
        if self.doubleClickCallback is not None:
            modelIndex = self.indexAt(event.pos())
            if modelIndex.isValid():
                self.doubleClickCallback(modelIndex)
        super().mouseDoubleClickEvent(event)


class FileChooser(QWidget):
    fileOpened = pyqtSignal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        # TODO: migrate to FolderComboBox?
        self.folderBox = QComboBox(self)
        self.explorerTree = FileTreeView(self)
        self.explorerTree.doubleClickCallback = self._fileOpened
        self.explorerModel = QFileSystemModel(self)
        self.explorerModel.setFilter(QDir.AllDirs | QDir.Files | QDir.NoDotAndDotDot)
        self.explorerModel.setNameFilters(["*.py"])
        self.explorerModel.setNameFilterDisables(False)
        self.explorerTree.setModel(self.explorerModel)
        for index in range(1, self.explorerModel.columnCount()):
            self.explorerTree.hideColumn(index)
        self.setCurrentFolder()
        self.folderBox.currentIndexChanged[int].connect(self.updateCurrentFolder)

        layout = QVBoxLayout(self)
        layout.addWidget(self.folderBox)
        layout.addWidget(self.explorerTree)
        layout.setContentsMargins(5, 5, 0, 0)

    def _fileOpened(self, modelIndex):
        path = self.explorerModel.filePath(modelIndex)
        if os.path.isfile(path):
            self.fileOpened.emit(path)

    def currentFolder(self):
        return self.explorerModel.rootPath()

    def setCurrentFolder(self, path=None):
        if path is None:
            app = QApplication.instance()
            path = app.getScriptsDirectory()
        else:
            assert os.path.isdir(path)
        self.explorerModel.setRootPath(path)
        self.explorerTree.setRootIndex(self.explorerModel.index(path))
        self.folderBox.blockSignals(True)
        self.folderBox.clear()
        style = self.style()
        dirIcon = style.standardIcon(style.SP_DirIcon)
        self.folderBox.addItem(dirIcon, os.path.basename(path))
        self.folderBox.insertSeparator(1)
        self.folderBox.addItem(self.tr("Browse…"))
        self.folderBox.setCurrentIndex(0)
        self.folderBox.blockSignals(False)

    def updateCurrentFolder(self, index):
        if index < self.folderBox.count() - 1:
            return
        path = QFileDialog.getExistingDirectory(
            self,
            self.tr("Choose Directory"),
            self.currentFolder(),
            QFileDialog.ShowDirsOnly,
        )
        if path:
            QSettings().setValue("scripting/path", path)
            self.setCurrentFolder(path)


class PythonEditor(BaseCodeEditor):
    openBlockDelimiter = ":"
    autocomplete = {
        Qt.Key_ParenLeft: ")",
        Qt.Key_BracketLeft: "]",
        Qt.Key_BraceLeft: "}",
        Qt.Key_Apostrophe: "",
        Qt.Key_QuoteDbl: "",
    }

    def __init__(self, parent=None):
        super().__init__(parent)
        self.highlighter = PythonHighlighter(self.document())

    def write(self, path):
        with open(path, "wt", encoding="utf-8") as outputFile:
            outputFile.write(self.toPlainText())
        self.document().setModified(False)

    def keyPressEvent(self, event):
        key = event.key()
        if key in self.autocomplete.keys():
            super().keyPressEvent(event)
            cursor = self.textCursor()
            cursor.insertText(self.autocomplete[key] or chr(key))
            cursor.movePosition(QTextCursor.PreviousCharacter)
            self.setTextCursor(cursor)
            return
        elif key == Qt.Key_Return:
            cursor = self.textCursor()
            ok = cursor.movePosition(QTextCursor.NextCharacter)
            if ok:
                cursor.movePosition(
                    QTextCursor.PreviousCharacter, QTextCursor.KeepAnchor, 2
                )
                combo = [
                    "{}{}".format(chr(k), v) for k, v in self.autocomplete.items() if v
                ]
                if not cursor.selectedText() in combo:
                    ok = False
            if ok:
                for _ in range(2):
                    super().keyPressEvent(event)
                cursor = self.textCursor()
                cursor.movePosition(QTextCursor.Up)
                cursor.insertText(self._indent)
                cursor.movePosition(QTextCursor.EndOfLine)
                self.setTextCursor(cursor)
                return
        elif key == Qt.Key_Backspace:
            cursor = self.textCursor()
            ok = cursor.movePosition(QTextCursor.PreviousCharacter)
            if ok:
                ok = cursor.movePosition(
                    QTextCursor.NextCharacter, QTextCursor.KeepAnchor, 2
                )
                tags = [
                    "{}{}".format(chr(k), v if v else chr(k))
                    for k, v in self.autocomplete.items()
                ]
                if ok and cursor.selectedText() in tags:
                    cursor.removeSelectedText()
                    event.accept()
                    return
        super().keyPressEvent(event)

    def dragEnterEvent(self, event):
        if event.source() != self:
            mimeData = event.mimeData()
            if mimeData.hasUrls():
                urls = mimeData.urls()
                for url in urls:
                    if url.isLocalFile():
                        event.acceptProposedAction()
                        break
                return
        super().dragEnterEvent(event)

    def dropEvent(self, event):
        if event.source() != self:
            mimeData = event.mimeData()
            if mimeData.hasUrls():
                urls = mimeData.urls()
                paths = [url.toLocalFile() for url in urls if url.isLocalFile()]
                text = ", ".join(paths)
                if len(paths) > 1:
                    text = "[%s]" % text
                textCursor = self.cursorForPosition(event.pos())
                textCursor.insertText(text)
                self.setTextCursor(textCursor)
                # HACK: Qt uses cleanup routines in a private namespace
                # called by the superclass. Ugh. We have no choice but to make
                # up an empty event if we want to exit drag state.
                mimeData = mimeData.__class__()
                event = event.__class__(
                    event.posF(),
                    event.possibleActions(),
                    mimeData,
                    event.mouseButtons(),
                    event.keyboardModifiers(),
                )
        super().dropEvent(event)

    def setPlainText(self, text):
        super().setPlainText(text)
        self.document().setModified(False)


class PythonHighlighter(BaseCodeHighlighter):
    def __init__(self, parent=None):
        super().__init__(parent)

        quotationFormat = QTextCharFormat()
        quotationFormat.setForeground(QColor(255, 27, 147))
        self.addBlockRule('u?r?"""', '"""', quotationFormat)
        self.addBlockRule("u?r?'''", "'''", quotationFormat)

        singleLineCommentFormat = QTextCharFormat()
        singleLineCommentFormat.setForeground(QColor(112, 128, 144))
        self.addRule("#[^\n]*", singleLineCommentFormat)

        quotationTemplate = "{0}[^{0}\\\\]*(\\\\.[^{0}\\\\]*)*{0}"
        self.addRule(
            "|".join(quotationTemplate.format(char) for char in ("'", '"')),
            quotationFormat,
        )

        classOrFnNameFormat = QTextCharFormat()
        classOrFnNameFormat.setForeground(QColor(96, 106, 161))
        self.addRule("(?<=\\bclass\\s|def\\s\\b)\\s*(\\w+)", classOrFnNameFormat)

        keywordFormat = QTextCharFormat()
        keywordFormat.setForeground(QColor(45, 95, 235))
        self.addRule("\\b(%s)\\b" % ("|".join(kwlist)), keywordFormat)


class ScriptingStatusBar(QStatusBar):
    def __init__(self, parent=None):
        super().__init__(parent)
        if platformSpecific.needsTighterMargins():
            margins = (6, -10, 9, -12)
        else:
            margins = (2, -1, 5, 0)
        self.setContentsMargins(*margins)
        self.setSizeGripEnabled(False)
        self.positionLabel = ClickLabel(self)
        self.indentLabel = IndentLabel(self)
        self.clearButton = QPushButton(self.tr("Clear"), self)
        self.runButton = QPushButton(self.tr("Run"), self)
        self.runButton.setShortcut("Ctrl+R")

        self.addWidget(self.positionLabel)
        self.addPermanentWidget(self.indentLabel)
        self.addPermanentWidget(self.clearButton)
        self.addPermanentWidget(self.runButton)

        self.indentModified = self.indentLabel.indentModified
        self.positionClicked = self.positionLabel.clicked
        self.clearButtonClicked = self.clearButton.clicked
        self.runButtonClicked = self.runButton.clicked

    def setIndent(self, indent):
        # TODO: this could be in a label widget subclass for better
        # encapsulation
        space = tab = 0
        for char in indent:
            if char == " ":
                space += 1
            elif char == "\t":
                tab += 1
        if space and not tab:
            if space < 2:
                text = self.tr("Space")
            else:
                text = self.tr("{}-Spaces").format(space)
        elif tab and not space:
            if tab < 2:
                text = self.tr("Tab")
            else:
                text = self.tr("Tabs")
        else:
            text = self.tr("Other")
        self.indentLabel.setText(text)

    def setPosition(self, textCursor):
        blockNumber, blockPosition = (
            textCursor.blockNumber(),
            textCursor.positionInBlock(),
        )
        # TODO: display selection information?
        self.positionLabel.setText("%d:%d" % (blockNumber + 1, blockPosition + 1))


class IndentLabel(ClickLabel):
    indentModified = pyqtSignal(str)

    def contextMenu(self):
        menu = QMenu(self)
        menu.addAction(self.tr("2-Spaces"), lambda: self.indentModified.emit("  "))
        menu.addAction(self.tr("4-Spaces"), lambda: self.indentModified.emit("    "))
        menu.addAction(self.tr("Tab"), lambda: self.indentModified.emit("\t"))
        return menu



from PyQt5.QtCore import QPoint, Qt, pyqtSignal
from PyQt5.QtWidgets import QLabel


class ClickLabel(QLabel):
    clicked = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setCursor(Qt.PointingHandCursor)

    def contextMenu(self):
        return None

    def showContextMenu(self):
        menu = self.contextMenu()
        if menu is None:
            return
        text = self.text()
        for action in menu.actions():
            if action.text() == text:
                action.setCheckable(True)
                action.setChecked(True)
        pos = QPoint(0, self.height())
        menu.exec_(self.mapToGlobal(pos))
        # leaveEvent isn't always triggered since we overlay a menu
        # cleanup manually.
        self._disableUnderline()

    def _disableUnderline(self):
        font = self.font()
        font.setUnderline(False)
        self.setFont(font)

    # ----------
    # Qt methods
    # ----------

    def enterEvent(self, event):
        font = self.font()
        font.setUnderline(True)
        self.setFont(font)

    def leaveEvent(self, event):
        self._disableUnderline()

    def mousePressEvent(self, event):
        self.showContextMenu()
        self.clicked.emit()



from PyQt5.QtCore import QEvent, QRect, QSize, Qt, pyqtSignal
from PyQt5.QtGui import QColor, QPainter, QPainterPath
from PyQt5.QtWidgets import QSizePolicy, QToolTip, QWidget

__all__ = ["TabWidget"]

sidePadding = 9
spacing = 2
topPadding = bottomPadding = 5
crossMargin = sidePadding
crossSize = 7

cross = QPainterPath()
cross.moveTo(0.38, 0.38)
cross.lineTo(9.63, 9.63)
cross.moveTo(0.38, 9.63)
cross.lineTo(9.63, 0.38)


class TabWidget(QWidget):
    """
    # TODO: RTL support?
    """

    currentTabChanged = pyqtSignal(int)
    tabRemoved = pyqtSignal(int)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setMouseTracking(True)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum)

        self._currentTab = None
        self._autoHide = True
        self._heroFirstTab = False
        self._hoverClose = None
        self._hoverTab = None
        self._tabs = []
        self._firstOffset = 0
        self._textWidth = 148

    def addTab(self, name, parent=None):
        self._tabs.append(name)
        self._recalcTextWidth()
        self._updateVisibility()
        self.update()

    def removeTab(self, index):
        del self._tabs[index]
        if self._currentTab >= index:
            if self._currentTab == index and self._currentTab < len(self._tabs):
                pass
            elif self._tabs:
                self._currentTab -= 1
            else:
                self._currentTab = None
            self.currentTabChanged.emit(self._currentTab)
        self.tabRemoved.emit(index)
        # XXX: somewhat hackish, we have no way to compute new rects except
        # post-repaint
        # idea: reuse rects but translate by removed tab width after tab index
        self._hoverClose = None
        self._recalcTextWidth()
        self._updateVisibility()
        self.update()

    def setTabName(self, index, name):
        self._tabs[index] = name
        self.update()

    def currentTab(self):
        if self._currentTab is None and self._tabs:
            self._currentTab = 0
        return self._currentTab

    def setCurrentTab(self, value):
        if value < 0:
            value = value % len(self._tabs)
        if self._currentTab == value:
            return
        self._currentTab = value
        self.currentTabChanged.emit(self._currentTab)
        self.update()

    def autoHide(self):
        return self._autoHide

    def setAutoHide(self, value):
        self._autoHide = value
        self._updateVisibility()

    def heroFirstTab(self):
        return self._heroFirstTab

    def setHeroFirstTab(self, value):
        if self._tabs:
            return
        self._heroFirstTab = value

    def tabs(self):
        return self._tabs

    def _recalcTextWidth(self):
        count = len(self._tabs)
        maxTextWidth = self.width() - count * (2 * sidePadding) - (count - 1) * spacing
        maxTextWidth -= (count - self._heroFirstTab) * (crossMargin + crossSize)
        oneWidth = maxTextWidth / count
        oneWidthInt = int(oneWidth)
        if oneWidthInt >= 148:
            self._textWidth = 148
            self._firstOffset = 0
        else:
            self._textWidth = oneWidthInt
            self._firstOffset = maxTextWidth - oneWidthInt * count

    def _updateVisibility(self):
        self.setVisible(len(self._tabs) > 1)

    # ----------
    # Qt methods
    # ----------

    def event(self, event):
        if event.type() == QEvent.ToolTip:
            index = None
            for recordIndex, rect in self._tabsRects.items():
                if QRect(*rect).contains(event.pos()):
                    index = recordIndex
            if index is not None:
                QToolTip.showText(event.globalPos(), self._tabs[index])
            else:
                QToolTip.hideText()
                event.ignore()
            return True
        return super().event(event)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            for recordIndex, rect in self._tabsRects.items():
                if QRect(*rect).contains(event.pos()):
                    self.setCurrentTab(recordIndex)
        else:
            super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if not hasattr(self, "_tabsRects"):
            return
        elements = [(self._closeRects, "_hoverClose"), (self._tabsRects, "_hoverTab")]
        changed = False
        for rects, attr in elements:
            selected = None
            for recordIndex, rect in rects.items():
                if QRect(*rect).contains(event.pos()):
                    selected = recordIndex
                    break
            if getattr(self, attr) != selected:
                setattr(self, attr, selected)
                changed = True
        if changed:
            self.update()

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            elements = [(self._closeRects, self.removeTab)]
            for rects, func in elements:
                for recordIndex, rect in rects.items():
                    if QRect(*rect).contains(event.pos()):
                        func(recordIndex)
                        return
        else:
            super().mouseReleaseEvent(event)

    def leaveEvent(self, event):
        elements = (self._hoverTab, self._hoverClose)
        if all(i is None for i in elements):
            return
        self._hoverTab = self._hoverClose = None
        self.update()

    def paintEvent(self, event):
        painter = QPainter(self)
        metrics = self.fontMetrics()

        # tabs
        self._closeRects = {}
        self._tabsRects = {}
        painter.save()
        currentTab = self.currentTab()
        tabFillColor = QColor(240, 240, 240)
        hoverFillColor = QColor(253, 253, 255)
        textHeight = metrics.lineSpacing()
        left = 0
        h = textHeight + topPadding * 2
        for index, name in enumerate(self.tabs()):
            isHeroFirstTab = not index and self._heroFirstTab
            isClosable = not isHeroFirstTab
            textWidth = self._textWidth
            if not index:
                textWidth += self._firstOffset
            w = sidePadding * 2 + textWidth
            if isClosable:
                w += crossMargin + crossSize
            self._tabsRects[index] = (left, 0, w, h)

            # background
            textColor = QColor(72, 72, 72)
            if index == currentTab:
                fillColor = tabFillColor
            elif index == self._hoverTab:
                fillColor = hoverFillColor
            else:
                fillColor = None
            if fillColor is not None:
                painter.fillRect(0, 0, w, h, fillColor)
            # text
            painter.save()
            painter.translate(sidePadding, metrics.ascent() + topPadding)
            painter.setPen(textColor)
            painter.setRenderHint(QPainter.Antialiasing)
            painter.drawText(0, 0, metrics.elidedText(name, Qt.ElideRight, textWidth))
            # cross
            if isClosable:
                # 3px padding for click rect
                self._closeRects[index] = (
                    left + textWidth + 2 * sidePadding - 3,
                    metrics.ascent() + topPadding - crossSize - 3,
                    crossSize + 6,
                    crossSize + 6,
                )
                if index == self._hoverClose:
                    color = QColor(254, 28, 28)
                else:
                    color = QColor(78, 78, 78)
                painter.setPen(color)
                pen = painter.pen()
                pen.setWidthF(1.5)
                painter.setPen(pen)
                painter.translate(textWidth + sidePadding, -crossSize)
                painter.setClipRect(0, 0, crossSize, crossSize)
                painter.scale(crossSize / 10, crossSize / 10)
                painter.drawPath(cross)
            painter.restore()
            # shift for the next tab
            shift = textWidth + 2 * sidePadding + spacing
            if isClosable:
                shift += crossMargin + crossSize
            painter.translate(shift, 0)
            left += shift
        painter.restore()

    def minimumSizeHint(self):
        height = topPadding + bottomPadding + self.fontMetrics().lineSpacing()
        return QSize(200, height)

    def sizeHint(self):
        width = 0
        metrics = self.fontMetrics()
        heroTab = self._heroFirstTab
        for name in self._tabs:
            width += metrics.width(name) + 2 * sidePadding + spacing
            if not heroTab:
                width += crossMargin + crossSize
            heroTab = False
        height = topPadding + bottomPadding + metrics.lineSpacing()
        return QSize(width, height)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self._recalcTextWidth()



import bisect

from PyQt5.QtCore import QSize, Qt, pyqtSignal
from PyQt5.QtGui import QKeySequence
from PyQt5.QtWidgets import QGridLayout, QPushButton, QRadioButton, QSizePolicy, QWidget

from defconQt.controls.glyphCellView import GlyphCellView, GlyphCellWidget
from defconQt.controls.listView import ListView
from defconQt.tools.glyphsMimeData import GlyphsMimeData
from trufont.controls.glyphStackWidget import GlyphStackWidget
from trufont.objects import icons, settings
from trufont.tools import platformSpecific

_leftGroupPrefix = "public.kern1"
_rightGroupPrefix = "public.kern2"


class GroupsWindow(QWidget):
    def __init__(self, font, parent=None):
        super().__init__(parent, Qt.Window)
        self._autoDirection = True
        self._font = font
        self._font.groups.addObserver(self, "_groupsChanged", "Groups.Changed")
        self._font.info.addObserver(self, "_fontInfoChanged", "Info.Changed")

        groups = self._font.groups
        self.groupsListView = GroupListView(self)
        self.groupsListView.setList(sorted(groups.keys()))
        self.groupsListView.setHeaderLabels(["Name"])
        self.groupsListView.alignmentChanged.connect(self._alignmentChanged)
        # self.groupsListView.groupDeleted.connect(self._groupDeleted)
        self.groupsListView.currentItemChanged.connect(self._groupChanged)
        self.groupsListView.valueChanged.connect(self._groupRenamed)

        self.stackWidget = GlyphStackWidget(self)

        self.addGroupButton = QPushButton(self)
        self.addGroupButton.setIcon(icons.i_plus())
        self.addGroupButton.setToolTip(self.tr("Add group"))
        self.addGroupButton.clicked.connect(self._groupAdd)
        self.removeGroupButton = QPushButton(self)
        self.removeGroupButton.setIcon(icons.i_minus())
        self.removeGroupButton.setToolTip(self.tr("Remove group"))
        self.removeGroupButton.clicked.connect(self._groupDeleted)
        if not groups:
            self.removeGroupButton.setEnabled(False)

        self.alignLeftBox = QRadioButton(self.tr("Align left"), self)
        self.alignRightBox = QRadioButton(self.tr("Align right"), self)
        self.alignRightBox.toggled.connect(self._alignmentChanged)
        self.alignLeftBox.setChecked(True)

        self.groupCellView = GroupCellView(font, self)
        self.groupCellView.glyphsDropped.connect(self._glyphsDropped)
        self.groupCellView.selectionDeleted.connect(self._selectionDeleted)

        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        layout = QGridLayout(self)
        layout.addWidget(self.groupsListView, 0, 0, 5, 3)
        layout.addWidget(self.stackWidget, 0, 3, 5, 4)
        layout.addWidget(self.addGroupButton, 5, 0)
        layout.addWidget(self.removeGroupButton, 5, 1)
        layout.addWidget(spacer, 5, 2)
        layout.addWidget(self.alignLeftBox, 5, 3)
        layout.addWidget(self.alignRightBox, 5, 6)
        layout.addWidget(self.groupCellView, 6, 0, 4, 7)
        layout.setColumnStretch(4, 1)
        self.setLayout(layout)

        self.updateWindowTitle(font=self._font)
        self.readSettings()

    def readSettings(self):
        geometry = settings.groupsWindowGeometry()
        if geometry:
            self.restoreGeometry(geometry)

    def writeSettings(self):
        settings.setGroupsWindowGeometry(self.saveGeometry())

    def updateWindowTitle(self, title=None, font=None):
        if title is None:
            title = self.tr("Groups")
        if font is not None:
            title = "{} – {} {}".format(
                title, font.info.familyName, font.info.styleName
            )
        self.setWindowTitle(title)

    # -------------
    # Notifications
    # -------------

    # font

    def _groupsChanged(self, notification):
        groups = notification.object
        self.groupsListView.setList(sorted(groups.keys()))
        name = self.groupsListView.currentValue()
        groupIsValid = name is not None and name in groups
        if groupIsValid:
            glyphs = [
                self._font[gName] for gName in groups[name] if gName in self._font
            ]
        else:
            glyphs = []
        self.groupCellView.setAcceptDrops(groupIsValid)
        self.stackWidget.setGlyphs(list(glyphs))
        self.groupCellView.setGlyphs(list(glyphs))

    def _fontInfoChanged(self, notification):
        self.updateWindowTitle(font=self._font)

    # widgets

    def _alignmentChanged(self, alignRight):
        self.stackWidget.setAlignment("right" if alignRight else "left")
        if alignRight:
            self.alignRightBox.blockSignals(True)
            self.alignRightBox.setChecked(True)
            self.alignRightBox.blockSignals(False)
        else:
            self.alignLeftBox.setChecked(True)

    def _groupAdd(self):
        groupName = self.tr("New group")
        groups = self._font.groups
        if groupName in groups:
            index = 1
            while "%s %d" % (groupName, index) in groups:
                index += 1
            groupName = "%s %d" % (groupName, index)
        groups.disableNotifications(observer=self)
        groups[groupName] = []
        # TODO; make this a listView method
        lst = self.groupsListView.list()
        index = bisect.bisect_left(lst, groupName)
        lst.insert(index, groupName)
        self.groupsListView.setList(lst)
        self.groupsListView.editRow(index)
        self.removeGroupButton.setEnabled(True)
        groups.enableNotifications(observer=self)

    def _groupChanged(self, name):
        if name is None:
            self.groupCellView.setAcceptDrops(False)
            return
        if self._autoDirection:
            if name.startswith(_leftGroupPrefix):
                self.alignRightBox.setChecked(True)
            if name.startswith(_rightGroupPrefix):
                self.alignLeftBox.setChecked(True)
        glyphs = []
        for gName in self._font.groups[name]:
            if gName in self._font:
                glyphs.append(self._font[gName])
        self.groupCellView.setAcceptDrops(True)
        self.stackWidget.setGlyphs(list(glyphs))
        self.groupCellView.setGlyphs(glyphs)

    def _groupRenamed(self, index, previous, current):
        if current == previous:
            return
        groups = self._font.groups
        groups.holdNotifications()
        groups[current] = groups[previous]
        del groups[previous]
        groups.releaseHeldNotifications()

    def _groupDeleted(self):
        name = self.groupsListView.currentValue()
        if name is None:
            return
        del self._font.groups[name]
        self.removeGroupButton.setEnabled(len(self._font.groups))

    def _selectionDeleted(self, selection):
        name = self.groupsListView.currentValue()
        if name is None:
            return
        groups = self._font.groups
        currentGroup = groups[name]
        # relying on ordered group elements
        # reverse to not change index of smaller elements
        for key in sorted(selection, reverse=True):
            del currentGroup[key]
        groups[name] = currentGroup
        groups.dirty = True

    def _glyphsDropped(self):
        name = self.groupsListView.currentValue()
        if name is None:
            return
        glyphs = self.groupCellView.glyphs()
        self._font.groups[name] = [glyph.name for glyph in glyphs]

    # ----------
    # Qt methods
    # ----------

    def sizeHint(self):
        return QSize(650, 650)

    def moveEvent(self, event):
        self.writeSettings()

    resizeEvent = moveEvent

    def closeEvent(self, event):
        super().closeEvent(event)
        if event.isAccepted():
            self._font.groups.removeObserver(self, "Groups.Changed")
            self._font.info.removeObserver(self, "Info.Changed")


class GroupListView(ListView):
    alignmentChanged = pyqtSignal(bool)
    groupDeleted = pyqtSignal(object)

    def currentValue(self):
        index = self.currentIndex()
        model = self.model()
        if model is None:
            return None
        return model.data(index)

    def editRow(self, index):
        self.editItem(index, 0)

    def keyPressEvent(self, event):
        key = event.key()
        if platformSpecific.isDeleteEvent(event):
            indexes = self.selectedIndexes()
            if indexes:
                data = self.model().data(indexes[0])
                self.groupDeleted.emit(data)
        elif key == Qt.Key_Left:
            self.alignmentChanged.emit(False)
        elif key == Qt.Key_Right:
            self.alignmentChanged.emit(True)
        else:
            super().keyPressEvent(event)


class GroupCellWidget(GlyphCellWidget):
    selectionDeleted = pyqtSignal(set)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._font = None

    def font_(self):
        return self._font

    def setFont_(self, font):
        self._font = font

    def keyPressEvent(self, event):
        if event.matches(QKeySequence.Delete):
            selection = self._selection
            if selection:
                self.selectionDeleted.emit(selection)
        else:
            super().keyPressEvent(event)

    def dragEnterEvent(self, event):
        mimeData = event.mimeData()
        if isinstance(mimeData, GlyphsMimeData):
            glyphs = mimeData.glyphs()
            for glyph in glyphs:
                if glyph.font == self._font:
                    event.acceptProposedAction()
                    return

    def dropEvent(self, event):
        mimeData = event.mimeData()
        # remove glyphs that aren't from our font or are already there
        glyphs = mimeData.glyphs()
        for index, glyph in enumerate(glyphs):
            if glyph.font != self._font or glyph in self._glyphs:
                del glyphs[index]
        mimeData.setGlyphs(glyphs)
        # now proceed
        super().dropEvent(event)


class GroupCellView(GlyphCellView):
    glyphCellWidgetClass = GroupCellWidget

    def __init__(self, font, parent=None):
        super().__init__(parent)
        self._glyphCellWidget.setFont_(font)
        # re-export signals
        self.selectionDeleted = self._glyphCellWidget.selectionDeleted

    def sizeHint(self):
        cellHeight = self._glyphCellWidget.cellSize()[1]
        size = super().sizeHint()
        size.setHeight(cellHeight * 2)
        return size



from collections.abc import Iterable

from defcon import Color
from PyQt5.QtCore import QByteArray, QSettings
from PyQt5.QtGui import QColor

from defconQt.tools.drawing import colorToQColor

_metricsWindowComboBoxItems = [
    "abcdefghijklmnopqrstuvwxyz",
    "ABCDEFGHIJKLMNOPQRSTUVWXYZ",
    "0123456789",
    "nn/? nono/? oo",
    "HH/? HOHO/? OO",
]

_latinDefaultName = "Latin-default"

_latinDefaultGlyphNames = [
    "space",
    "exclam",
    "quotesingle",
    "quotedbl",
    "numbersign",
    "dollar",
    "percent",
    "ampersand",
    "parenleft",
    "parenright",
    "asterisk",
    "plus",
    "comma",
    "hyphen",
    "period",
    "slash",
    "zero",
    "one",
    "two",
    "three",
    "four",
    "five",
    "six",
    "seven",
    "eight",
    "nine",
    "colon",
    "semicolon",
    "less",
    "equal",
    "greater",
    "question",
    "at",
    "A",
    "B",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "I",
    "J",
    "K",
    "L",
    "M",
    "N",
    "O",
    "P",
    "Q",
    "R",
    "S",
    "T",
    "U",
    "V",
    "W",
    "X",
    "Y",
    "Z",
    "bracketleft",
    "backslash",
    "bracketright",
    "asciicircum",
    "underscore",
    "grave",
    "a",
    "b",
    "c",
    "d",
    "e",
    "f",
    "g",
    "h",
    "i",
    "j",
    "k",
    "l",
    "m",
    "n",
    "o",
    "p",
    "q",
    "r",
    "s",
    "t",
    "u",
    "v",
    "w",
    "x",
    "y",
    "z",
    "braceleft",
    "bar",
    "braceright",
    "asciitilde",
    "exclamdown",
    "cent",
    "sterling",
    "currency",
    "yen",
    "brokenbar",
    "section",
    "copyright",
    "ordfeminine",
    "guillemotleft",
    "logicalnot",
    "registered",
    "macron",
    "degree",
    "plusminus",
    "twosuperior",
    "threesuperior",
    "mu",
    "paragraph",
    "periodcentered",
    "onesuperior",
    "ordmasculine",
    "guillemotright",
    "onequarter",
    "onehalf",
    "threequarters",
    "questiondown",
    "Agrave",
    "Aacute",
    "Acircumflex",
    "Atilde",
    "Adieresis",
    "Aring",
    "AE",
    "Ccedilla",
    "Egrave",
    "Eacute",
    "Ecircumflex",
    "Edieresis",
    "Igrave",
    "Iacute",
    "Icircumflex",
    "Idieresis",
    "Eth",
    "Ntilde",
    "Ograve",
    "Oacute",
    "Ocircumflex",
    "Otilde",
    "Odieresis",
    "multiply",
    "Oslash",
    "Ugrave",
    "Uacute",
    "Ucircumflex",
    "Udieresis",
    "Yacute",
    "Thorn",
    "germandbls",
    "agrave",
    "aacute",
    "acircumflex",
    "atilde",
    "adieresis",
    "aring",
    "ae",
    "ccedilla",
    "egrave",
    "eacute",
    "ecircumflex",
    "edieresis",
    "igrave",
    "iacute",
    "icircumflex",
    "idieresis",
    "eth",
    "ntilde",
    "ograve",
    "oacute",
    "ocircumflex",
    "otilde",
    "odieresis",
    "divide",
    "oslash",
    "ugrave",
    "uacute",
    "ucircumflex",
    "udieresis",
    "yacute",
    "thorn",
    "ydieresis",
    "dotlessi",
    "gravecomb",
    "acutecomb",
    "uni0302",
    "uni0308",
    "uni030A",
    "tildecomb",
    "uni0327",
    "quoteleft",
    "quoteright",
    "minus",
]

_fallbackValues = {
    "export/removeOverlap": False,  # XXX: make these true when implemented
    "export/autohint": False,
    "export/useDirectory": True,
    "fontWindow/glyphCellSize": 86,
    "fontWindow/propertiesHidden": False,
    "metricsWindow/comboBoxItems": _metricsWindowComboBoxItems,
    "misc/loadRecentFile": False,
    "outputWindow/wrapLines": False,
    "scriptingWindow/hSplitterSizes": [0, 1],
    "scriptingWindow/vSplitterSizes": [1, 100],
    "settings/defaultGlyphSet": _latinDefaultName,
}

# --------------------------
# Generic get/set + fallback
# --------------------------

_type = type


def value(key, fallback=None, type=None):
    settings = QSettings()
    if fallback is None and key in _fallbackValues:
        fallback = _fallbackValues[key]
        if isinstance(fallback, Iterable):
            type = _type(fallback[0])
        else:
            type = _type(fallback)
    return settings.value(key, fallback, type)


def setValue(key, value):
    # TODO: send preferencesChanged here?
    settings = QSettings()
    settings.setValue(key, value)


# -----------
# Convenience
# -----------

# value

# XXX: use more robust setting
# defaultGlyphSet will always be the first in array and we have a bool
# to say whether we should apply it


def fontWindowGeometry():
    return value("fontWindow/geometry", type=QByteArray)


def setFontWindowGeometry(geometry):
    setValue("fontWindow/geometry", geometry)


def fontFeaturesWindowGeometry():
    return value("fontFeaturesWindow/geometry", type=QByteArray)


def setFontFeaturesWindowGeometry(geometry):
    setValue("fontFeaturesWindow/geometry", geometry)


def fontInfoWindowGeometry():
    return value("fontInfoWindow/geometry", type=QByteArray)


def setFontInfoWindowGeometry(geometry):
    setValue("fontInfoWindow/geometry", geometry)


def glyphCellSize():
    return value("fontWindow/glyphCellSize")


def setGlyphCellSize(cellSize):
    setValue("fontWindow/glyphCellSize", cellSize)


def removeGlyphCellSize():
    settings = QSettings()
    settings.remove("fontWindow/glyphCellSize")


def propertiesHidden():
    return value("fontWindow/propertiesHidden")


def setPropertiesHidden(value):
    setValue("fontWindow/propertiesHidden", value)


def drawingAttributes():
    return value("fontWindow/drawingAttributes", type=dict)


def setDrawingAttributes(drawingAttributes):
    setValue("fontWindow/drawingAttributes", drawingAttributes)


def groupsWindowGeometry():
    return value("groupsWindow/geometry", type=QByteArray)


def setGroupsWindowGeometry(geometry):
    setValue("groupsWindow/geometry", geometry)


def kerningWindowGeometry():
    return value("kerningWindow/geometry", type=QByteArray)


def setKerningWindowGeometry(geometry):
    setValue("kerningWindow/geometry", geometry)


def metricsWindowGeometry():
    return value("metricsWindow/geometry", type=QByteArray)


def setMetricsWindowGeometry(geometry):
    setValue("metricsWindow/geometry", geometry)


def outputWindowGeometry():
    return value("outputWindow/geometry", type=QByteArray)


def setOutputWindowGeometry(geometry):
    setValue("outputWindow/geometry", geometry)


def outputWindowWrapLines():
    return value("outputWindow/wrapLines")


def setOutputWindowWrapLines(value):
    setValue("outputWindow/wrapLines", value)


def scriptingWindowGeometry():
    return value("scriptingWindow/geometry", type=QByteArray)


def setScriptingWindowGeometry(geometry):
    setValue("scriptingWindow/geometry", geometry)


def scriptingWindowHSplitterSizes():
    return value("scriptingWindow/hSplitterSizes")


def setScriptingWindowHSplitterSizes(sizes):
    setValue("scriptingWindow/hSplitterSizes", sizes)


def scriptingWindowVSplitterSizes():
    return value("scriptingWindow/vSplitterSizes", type=int)


def setScriptingWindowVSplitterSizes(sizes):
    setValue("scriptingWindow/vSplitterSizes", sizes)


def settingsWindowGeometry():
    return value("settingsWindow/geometry", type=QByteArray)


def setSettingsWindowGeometry(geometry):
    setValue("settingsWindow/geometry", geometry)


def defaultGlyphSet():
    return value("settings/defaultGlyphSet")


def setDefaultGlyphSet(name):
    if name is None:
        name = ""
    setValue("settings/defaultGlyphSet", name)


def glyphListPath():
    return value("settings/glyphListPath", type=str)


def setGlyphListPath(path):
    if path is None:
        path = ""
    setValue("settings/glyphListPath", path)


def removeGlyphListPath(path):
    settings = QSettings()
    settings.remove("settings/glyphListPath")


def metricsWindowComboBoxItems():
    return value("metricsWindow/comboBoxItems")


def setMetricsWindowComboBoxItems(items):
    setValue("metricsWindow/comboBoxItems", items)


def importFileDialogState():
    return value("core/importFileDialogState", type=QByteArray)


def setImportFileDialogState(state):
    setValue("core/importFileDialogState", state)


def openFileDialogState():
    return value("core/openFileDialogState", type=QByteArray)


def setOpenFileDialogState(state):
    setValue("core/openFileDialogState", state)


def saveFileDialogState():
    return value("core/saveFileDialogState", type=QByteArray)


def setSaveFileDialogState(state):
    setValue("core/saveFileDialogState", state)


def exportFileDialogState():
    return value("export/fileDialogState", type=QByteArray)


def setExportFileDialogState(state):
    setValue("export/fileDialogState", state)


def exportFileFormats():
    return value("export/fileFormats", type=str)


def setExportFileFormats(formats):
    setValue("export/fileFormats", formats)


def exportCompressionFormats():
    return value("export/compressionFormats", type=str)


def setExportCompressionFormats(formats):
    setValue("export/compressionFormats", formats)


def exportRemoveOverlap():
    return value("export/removeOverlap")


def setExportRemoveOverlap(value):
    setValue("export/removeOverlap", value)


def exportAutohint():
    return value("export/autohint")


def setExportAutohint(value):
    setValue("export/autohint", value)


def exportUseDirectory():
    return value("export/useDirectory")


def setExportUseDirectory(value):
    setValue("export/useDirectory", value)


def exportDirectory():
    return value("export/directory", type=str)


def setExportDirectory(value):
    setValue("export/directory", value)


def scriptingFileDialogState():
    return value("scriptingWindow/fileDialogState", type=QByteArray)


def setScriptingFileDialogState(state):
    setValue("scriptingWindow/fileDialogState", state)


def loadRecentFile():
    return value("misc/loadRecentFile")


def setLoadRecentFile(value):
    setValue("misc/loadRecentFile", value)


def recentFiles():
    return value("core/recentFiles", [], type=list)


def setRecentFiles(recentFiles):
    setValue("core/recentFiles", recentFiles)


# containers


def readGlyphSets():
    settings = QSettings()
    size = settings.beginReadArray("glyphSets")
    glyphSets = {}
    if not size:
        glyphSets[_latinDefaultName] = _latinDefaultGlyphNames
    for i in range(size):
        settings.setArrayIndex(i)
        glyphSetName = settings.value("name", type=str)
        glyphSetGlyphNames = settings.value("glyphNames", type=str)
        glyphSets[glyphSetName] = glyphSetGlyphNames
    settings.endArray()
    return glyphSets


def writeGlyphSets(glyphSets):
    settings = QSettings()
    settings.beginWriteArray("glyphSets", len(glyphSets))
    index = 0
    for name, cset in glyphSets.items():
        settings.setArrayIndex(index)
        settings.setValue("name", name)
        settings.setValue("glyphNames", cset)
        index += 1
    settings.endArray()


# mark colors


def readMarkColors():
    settings = QSettings()
    size = settings.beginReadArray("misc/markColors")
    if not size:
        markColors = [
            [QColor(255, 0, 0), "Red"],
            [QColor(255, 255, 0), "Yellow"],
            [QColor(0, 255, 0), "Green"],
        ]
    else:
        markColors = []
    for i in range(size):
        settings.setArrayIndex(i)
        markColor = settings.value("color", type=str)
        markColorName = settings.value("name", type=str)
        markColors.append([colorToQColor(markColor), markColorName])
    settings.endArray()
    return markColors


def writeMarkColors(markColors):
    settings = QSettings()
    settings.beginWriteArray("misc/markColors")
    # serialized in UFO form
    i = 0
    for color, name in markColors:
        settings.setArrayIndex(i)
        settings.setValue("color", str(Color(color.getRgbF())))
        settings.setValue("name", name)
        i += 1
    settings.endArray()



"""
The *listView* submodule
------------------------

The *listView* submodule provides a widget that can conveniently display Python
lists_.

.. _lists: https://docs.python.org/3/tutorial/introduction.html#lists
"""

import collections.abc

from defcon import Font, Glyph
from PyQt5.QtCore import QAbstractTableModel, QModelIndex, Qt, pyqtSignal
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QCheckBox,
    QColorDialog,
    QProxyStyle,
    QStyle,
    QStyledItemDelegate,
    QStyleOption,
    QTreeView,
)

from defconQt.tools import drawing

__all__ = ["ListView"]


class AbstractListModel(QAbstractTableModel):
    valueChanged = pyqtSignal(QModelIndex, object, object)

    def __init__(self, lst, parent=None):
        super().__init__(parent)
        self._headerLabels = []
        self._inDrop = False

        self.setList(lst)

    # implementation-specific methods

    def _data(self, row, column):
        raise NotImplementedError

    def _setData(self, row, column, value):
        raise NotImplementedError

    def _columnCount(self):
        raise NotImplementedError

    def _rowCount(self):
        raise NotImplementedError

    def _insertRows(self):
        raise NotImplementedError

    def _removeRows(self):
        raise NotImplementedError

    # other methods

    def list(self):
        return list(self._list)

    def setList(self, lst):
        self.layoutAboutToBeChanged.emit()
        self._list = lst
        self.layoutChanged.emit()

    def headerLabels(self):
        return list(self._headerLabels)

    def setHeaderLabels(self, labels):
        self._headerLabels = labels

    def sorted(self):
        return self._sorted

    def setSorted(self, value):
        self._sorted = value
        if value:
            self._list = sorted(self._list)

    # builtins

    def data(self, index, role=Qt.DisplayRole):
        if index.isValid() and role in (Qt.DisplayRole, Qt.EditRole):
            row, column = index.row(), index.column()
            return self._data(row, column)
        return None

    def setData(self, index, value, role=Qt.EditRole):
        if index.isValid() and role in (Qt.DisplayRole, Qt.EditRole):
            row, column = index.row(), index.column()
            oldValue = self._data(row, column)
            self._setData(row, column, value)
            if not self._inDrop:
                self.valueChanged.emit(index, oldValue, value)
            self.dataChanged.emit(index, index, [role])
            return True
        return super().setData(index, value, role)

    def columnCount(self, parent=QModelIndex()):
        # flat table
        # http://stackoverflow.com/a/27333368/2037879
        if parent.isValid():
            return 0
        return self._columnCount()

    def rowCount(self, parent=QModelIndex()):
        if parent.isValid():
            return 0
        return self._rowCount()

    def insertRows(self, row, count, parent=QModelIndex()):
        if parent.isValid():
            return False
        self.beginInsertRows(parent, row, row + count - 1)
        self._insertRows(row, count)
        self.endInsertRows()
        return True

    def removeRows(self, row, count, parent=QModelIndex()):
        if parent.isValid():
            return False
        self.beginRemoveRows(parent, row, row + count - 1)
        self._removeRows(row, count)
        self.endRemoveRows()
        return True

    def dropMimeData(self, data, action, row, column, parent):
        if parent.isValid():
            return False
        # force column 0, we want to drag column onto the start of the
        # drop spot, otherwise it will be split between two new columns
        self._inDrop = True
        result = super().dropMimeData(data, action, row, 0, parent)
        self._inDrop = False
        return result

    def flags(self, index):
        flags = (
            Qt.ItemIsEnabled
            | Qt.ItemIsSelectable
            | Qt.ItemIsDragEnabled
            | Qt.ItemIsDropEnabled
            | Qt.ItemNeverHasChildren
        )
        if not isinstance(self.data(index), (Font, Glyph)):
            flags |= Qt.ItemIsEditable
        return flags

    def supportedDropActions(self):
        return Qt.CopyAction | Qt.MoveAction

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            if section >= len(self._headerLabels):
                return None
            else:
                return self._headerLabels[section]
        return super().headerData(section, orientation, role)


class FlatListModel(AbstractListModel):
    def __init__(self, lst, columnCount=1, parent=None):
        super().__init__(lst, parent)
        self._columns = columnCount

    def _data(self, row, column):
        return self._list[row * self._columns + column]

    def _setData(self, row, column, value):
        self._list[row * self._columns + column] = value

    def _columnCount(self):
        return self._columns

    def _rowCount(self):
        return len(self._list) // self._columns

    def _insertRows(self, row, count):
        # XXX: test this!
        for index in range(row, row + count):
            for _ in self._columns:
                self._list.insert(index, 0)

    def _removeRows(self, row, count):
        # XXX: test this!
        for index in range(row, row + count):
            for _ in self._columns:
                del self._list[index * self._columns]


class OneTwoListModel(AbstractListModel):
    def _data(self, row, column):
        if self._is2D:
            return self._list[row][column]
        else:
            assert column == 0
            return self._list[row]

    def _setData(self, row, column, value):
        if self._is2D:
            self._list[row][column] = value
        else:
            assert column == 0
            self._list[row] = value

    def _columnCount(self):
        if self._is2D:
            if self._list:
                return len(self._list[0])
            return 0
        else:
            return 1

    def _rowCount(self):
        return len(self._list)

    def _insertRows(self, row, count):
        for index in range(row, row + count):
            elem = [0] * self._columnCount() if self._is2D else 0
            self._list.insert(index, elem)

    def _removeRows(self, row, count):
        for index in range(row, row + count):
            del self._list[index]

    def setList(self, lst):
        self.layoutAboutToBeChanged.emit()
        self._list = lst
        if self._list and isinstance(self._list[0], collections.abc.MutableSequence):
            self._is2D = True
        else:
            self._is2D = False
        self.layoutChanged.emit()


class ListItemDelegate(QStyledItemDelegate):
    def createEditor(self, parent, option, index):
        data = index.data(Qt.EditRole)
        if isinstance(data, bool):
            checkBox = QCheckBox(parent)
            checkBox.setChecked(data)
            checkBox.toggled.connect(
                lambda: self.setModelData(checkBox, index.model(), index)
            )
            return checkBox
        elif isinstance(data, QColor):
            # we have our own ways
            return None
        return super().createEditor(parent, option, index)

    def displayText(self, value, locale):
        if isinstance(value, Font):
            info = value.info
            return f"{info.familyName} {info.styleName}"
        elif isinstance(value, Glyph):
            return value.name
        elif isinstance(value, (QColor, bool)):
            # we'll paint those instead
            return None
        return super().displayText(value, locale)

    def paint(self, painter, option, index):
        super().paint(painter, option, index)
        data = index.data(Qt.DisplayRole)
        if isinstance(data, QColor):
            rect = option.rect.adjusted(2, 1, -2, -1)
            if data.isValid():
                painter.fillRect(rect, data)
            else:
                backgroundColor = QColor(214, 214, 214)
                color = QColor(170, 170, 170)
                tileSize = round(rect.height() / 3)
                drawing.drawTiles(
                    painter,
                    rect,
                    tileSize=tileSize,
                    color=color,
                    backgroundColor=backgroundColor,
                )

    def setModelData(self, editor, model, index):
        data = index.data(Qt.EditRole)
        if isinstance(data, bool):
            value = editor.isChecked()
            model.setData(index, value)
        else:
            super().setModelData(editor, model, index)


class ListProxy(QProxyStyle):
    def drawPrimitive(self, element, option, painter, widget):
        # http://stackoverflow.com/a/9611137/2037879
        if element == QStyle.PE_IndicatorItemViewItemDrop and not option.rect.isNull():
            # we don't drag over items
            # XXX: possible to forbid this in the model instead?
            if option.rect.height():
                return
            option_ = QStyleOption(option)
            option_.rect.setLeft(0)
            if widget is not None:
                option_.rect.setRight(widget.width())
            super().drawPrimitive(element, option_, painter, widget)
        else:
            super().drawPrimitive(element, option, painter, widget)


class ListView(QTreeView):
    """
    A QTreeView_ widget that displays a Python list, whether 1D or 2D.

    Besides standard types, this widget can display Font_ or Glyph_.

    Use *setDragEnabled(True)* to allow reordering drag and drop.

    Emits *listChanged* when data changes inside the widget (when performing
    drag and drop, mostly).

    # TODO: cleanup API and compare w QTreeWidget
    # TODO: maybe clear widgets on setList() and try to do without setIndexWidget
    # TODO: make it possible to up/down selected row w shortcut
    # e.g. Alt+Up/Down

    .. _Font: http://ts-defcon.readthedocs.org/en/ufo3/objects/font.html
    .. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
    .. _QTreeView: http://doc.qt.io/qt-5/qtreeview.html
    """

    currentItemChanged = pyqtSignal(object)
    selectionChanged_ = pyqtSignal()

    flatListModelClass = FlatListModel
    oneTwoListModelClass = OneTwoListModel

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setDragDropMode(QAbstractItemView.InternalMove)
        self.setDragEnabled(False)
        self.setItemDelegate(ListItemDelegate())
        self.setRootIsDecorated(False)
        self.setStyle(ListProxy())
        self.header().setVisible(False)
        self.doubleClicked.connect(self._doubleClicked)
        self._flatListInput = False
        self._triggers = None

    def _doubleClicked(self, index):
        model = self.model()
        if model is None:
            return
        data = model.data(index)
        if isinstance(data, QColor):
            if QApplication.keyboardModifiers() & Qt.AltModifier:
                model.setData(index, QColor())
            else:
                dialog = QColorDialog(self)
                dialog.setCurrentColor(data)
                dialog.setOption(QColorDialog.ShowAlphaChannel)
                ret = dialog.exec_()
                if ret:
                    color = dialog.currentColor()
                    model.setData(index, color)

    # -------------------
    # Convenience methods
    # -------------------

    def currentRow(self):
        index = self.currentIndex()
        data = self.model().list()
        return data[index.row()]

    def removeCurrentRow(self):
        index = self.currentIndex()
        model = self.model()
        model.removeRow(index.row())

    def editItem(self, row, column):
        model = self.model()
        if model is None:
            return
        self.edit(model.index(row, column))

    def setCurrentItem(self, row, column):
        model = self.model()
        if model is None:
            return
        index = model.index(row, column)
        super().setCurrentIndex(index)

    def selectedRows(self):
        selectionModel = self.selectionModel()
        if selectionModel is None:
            return []
        selectedRows = selectionModel.selectedRows()
        return [index.row() for index in selectedRows]

    def setEditable(self, value):
        """
        Sets whether the list’s elements can be edited.
        """
        if value:
            if self._triggers is None:
                return
            # default actions vary depending on platform
            self.setEditTriggers(self._triggers)
        else:
            self._triggers = self.editTriggers()
            self.setEditTriggers(QAbstractItemView.NoEditTriggers)

    def headerLabels(self):
        """
        Returns this widget’s current header labels.
        """
        model = self.model()
        if model is None:
            return None
        return model.headerLabels()

    def setHeaderLabels(self, labels):
        """
        Sets the header labels to *labels* (should be a list of strings).
        """
        model = self.model()
        if model is None:
            return
        model.setHeaderLabels(labels)
        self.header().setVisible(bool(labels))

    def list(self):
        """
        Returns the list as displayed by this widget, or None if no list was
        specified.
        """
        model = self.model()
        if model is None:
            return None
        return model.list()

    def setList(self, lst, **kwargs):
        """
        Sets the widget to display list *lst*.

        Additional keyword arguments may be provided and forwarded to the
        model.

        The default is None.

        # TODO: we should maybe clear indexWidgets here
        """
        model = self.model()
        if model is None:
            if self._flatListInput:
                modelClass = self.flatListModelClass
            else:
                modelClass = self.oneTwoListModelClass
            model = modelClass(lst, **kwargs)
            self.dataDropped = model.rowsRemoved
            self.valueChanged = model.valueChanged
            self.setModel(model)
        else:
            index = self.currentIndex()
            # reset the selection model to avoid selected row out of range
            # http://stackoverflow.com/a/15878679/2037879
            self.selectionModel().reset()
            model.setList(lst)
            if index.row() < model.rowCount() and index.column() < model.columnCount():
                self.setCurrentIndex(index)

    def flatListInput(self):
        """
        Returns whether this widget takes a flat list as input.
        """
        return self._flatListInput

    def setFlatListInput(self, value):
        """
        Sets whether :func:`setList` should consider its input list as flat,
        i.e. as a 2D structure.
        In that case, pass the *columnCount* argument to specify how many
        columns should be considered, e.g.:

        >>> from defconQt.controls.listView import ListView
        >>> view = ListView()
        >>> view.setList(
        ...     [
        ...         [1, 2, 3],
        ...         [4, 5, 6]
        ...     ])

        is equivalent to:

        >>> view = ListView()
        >>> view.setFlatList(True)
        >>> view.setList([1, 2, 3, 4, 5, 6], columnCount=3)

        """
        self._flatListInput = value

    # ----------
    # Qt methods
    # ----------

    def currentChanged(self, current, previous):
        super().currentChanged(current, previous)
        model = self.model()
        self.currentItemChanged.emit(model.data(current))

    def selectionChanged(self, selected, deselected):
        super().selectionChanged(selected, deselected)
        self.selectionChanged_.emit()

    def dropEvent(self, event):
        if event.source() == self:
            # widgets bookkeeping
            cachedWidgets = []
            # figure out the indexes
            dragRow = self.currentIndex().row()
            dropRow = self.indexAt(event.pos()).row()
            if self.dropIndicatorPosition() == QAbstractItemView.BelowItem:
                dropRow += 1
            # extract
            model = self.model()
            for col in range(model.columnCount()):
                widget = self.indexWidget(model.index(dragRow, 0))
                if widget:
                    # release the widget
                    self.editorDestroyed(widget)
                    # store it
                    cachedWidgets.append((col, widget))
            super().dropEvent(event)
            for col, widget in cachedWidgets:
                index = model.index(dropRow, col)
                self.setIndexWidget(index, widget)
        else:
            super().dropEvent(event)
